#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
File consists of helping functions and variables.

The functions and variables are utilised in the converting tool
to convert from nyaml to nxdl and vice versa.
"""

import hashlib
from typing import Callable

from yaml.composer import Composer
from yaml.constructor import Constructor
from yaml.loader import Loader
from yaml.nodes import ScalarNode
from yaml.resolver import BaseResolver

# Yaml library does not except the keys (escape char "\t" and yaml separator ":")
ESCAPE_CHAR_DICT_IN_YAML = {"\t": "    "}
ESCAPE_CHAR_DICT_IN_XML = {val: key for key, val in ESCAPE_CHAR_DICT_IN_YAML.items()}

# Set up attributes for nxdl version
NXDL_GROUP_ATTRIBUTES = (
    "optional",
    "recommended",
    "name",
    "type",
    "maxOccurs",
    "minOccurs",
    "deprecated",
)
NXDL_FIELD_ATTRIBUTES = (
    "optional",
    "recommended",
    "name",
    "type",
    "axes",
    "axis",
    "data_offset",
    "interpretation",
    "long_name",
    "maxOccurs",
    "minOccurs",
    "nameType",
    "primary",
    "signal",
    "stride",
    "required",
    "deprecated",
    "units",
)

NXDL_ATTRIBUTES_ATTRIBUTES = (
    "name",
    "type",
    "recommended",
    "optional",
    "deprecated",
)

NXDL_LINK_ATTRIBUTES = ("name", "target", "napimount")

# Set up attributes for yaml version
YAML_GROUP_ATTRIBUTES = (*NXDL_GROUP_ATTRIBUTES, "exists")

YAML_FIELD_ATTRIBUTES = (*NXDL_FIELD_ATTRIBUTES[0:-1], "unit", "exists", "dim")

YAML_ATTRIBUTES_ATTRIBUTES = (
    *NXDL_ATTRIBUTES_ATTRIBUTES,
    "minOccurs",
    "maxOccurs",
    "exists",
)

YAML_LINK_ATTRIBUTES = NXDL_LINK_ATTRIBUTES


def remove_namespace_from_tag(tag):
    """Helper function to remove the namespace from an XML tag."""
    if isinstance(tag, Callable) and tag.__name__ == "Comment":
        return "!--"
    return tag.split("}")[-1]


class LineLoader(Loader):  # pylint: disable=too-many-ancestors
    """Class to load yaml file with extra non yaml items.

    LineLoader parses a yaml into a python dictionary extended with extra items.
    The new items have as keys __line__<yaml_keyword> and as values the yaml file line number
    """

    def compose_node(self, parent, index):
        """Compose node and return node."""
        # the line number where the previous token has ended (plus empty lines)
        node = Composer.compose_node(self, parent, index)
        node.__line__ = self.line + 1
        return node

    def construct_mapping(self, node, deep=False):
        """Construct mapping between node info and line info."""
        node_pair_lst_for_appending = []

        # Visit through node-pair list
        for key_node in node.value:
            shadow_key_node = ScalarNode(
                tag=BaseResolver.DEFAULT_SCALAR_TAG,
                value="__line__" + key_node[0].value,
            )
            shadow_value_node = ScalarNode(
                tag=BaseResolver.DEFAULT_SCALAR_TAG, value=key_node[0].__line__
            )
            node_pair_lst_for_appending.append((shadow_key_node, shadow_value_node))

        node.value = node.value + node_pair_lst_for_appending
        return Constructor.construct_mapping(self, node, deep=deep)


def get_yaml_escape_char_dict():
    """Get escape char and the way to skip them in yaml."""
    return ESCAPE_CHAR_DICT_IN_YAML


def get_yaml_escape_char_reverter_dict():
    """To revert yaml escape char in xml constructor from yaml."""

    return ESCAPE_CHAR_DICT_IN_XML


def type_check(nx_type):
    """Check for nexus type if type is NX_CHAR get '' or get as it is."""

    if nx_type in ["NX_CHAR", ""]:
        nx_type = ""
    else:
        nx_type = f"({nx_type})"
    return nx_type


def get_node_parent_info(tree, node):
    """Return tuple of (parent, index).

    parent = parent node is the first level node under tree node
    index = index of grand child node of tree
    """

    # map from grand child to parent which is child of tree
    parent_map = {c: p for p in tree.iter() for c in p}
    parent = parent_map[node]
    return parent, list(parent).index(node)


def clean_empty_lines(line_list):
    """Clean up empty lines by top part and bottom and part."""
    if not isinstance(line_list, list):
        line_list = line_list.split("\n") if "\n" in line_list else [""]

    start_non_empty_line = -1
    ends_non_empty_line = None
    # Find the index of first non-empty line
    for ind, line in enumerate(line_list):
        if len(line.strip()) > 1:
            start_non_empty_line = ind
            break

    # Find the index of the last non-empty line
    for ind, line in enumerate(reversed(line_list)):
        if len(line.strip()) > 1:
            ends_non_empty_line = -ind
            break

    if ends_non_empty_line == 0:
        ends_non_empty_line = None
    return line_list[start_non_empty_line:ends_non_empty_line]


def nx_name_type_resolving(tmp):
    """Separate name and NeXus type

    Extracts the eventual custom name {optional_string}
    and type {nexus_type} from a YML section string.
    YML section string syntax: optional_string(nexus_type)
    """
    if tmp.count("(") == 1 and tmp.count(")") == 1:
        # we can safely assume that every valid YML key resolves
        # either an nx_ (type, base, candidate) class contains only 1 '(' and ')'
        index_start = tmp.index("(")
        index_end = tmp.index(")", index_start + 1)
        if index_start > index_end:
            raise ValueError(
                f"Check name and type combination {tmp} which can not be resolved."
            )
        if index_end - index_start == 1:
            raise ValueError(
                f"Check name(type) combination {tmp}, properly not defined."
            )
        typ = tmp[index_start + 1 : index_end]
        nam = tmp.replace("(" + typ + ")", "")
        return nam, typ

    # or a name for a member
    typ = ""
    nam = tmp
    return nam, typ


def get_sha256_hash(file_name):
    """Generate a sha256_hash for a given file."""
    sha_hash = hashlib.sha256()

    with open(
        file=file_name,
        mode="rb",
    ) as file_obj:
        # Update hash for each 4k block of bytes
        for b_line in iter(lambda: file_obj.read(4096), b""):
            sha_hash.update(b_line)
    return sha_hash.hexdigest()


def extend_yamlfile_by_nxdl_as_comment(
    yaml_file, file_to_be_appended, top_lines_list=None
):
    """Extend yaml file by the file_to_be_appended as comment."""

    with open(yaml_file, mode="a+", encoding="utf-8") as f1_obj:
        if top_lines_list:
            for line in top_lines_list:
                f1_obj.write(line)

        with open(file_to_be_appended, mode="r", encoding="utf-8") as f2_obj:
            for line in f2_obj:
                f1_obj.write(f"# {line}")


def separate_hash_yaml_and_nxdl(yaml_file, sep_yaml, sep_xml):
    """Separate yaml, SHA hash and nxdl parts.

    Separate the provided yaml file into yaml, nxdl and hash if yaml was extended with
    nxdl at the end of yaml as

                    <yaml part>
        '\n# ++++++++++++++++++++++++++++++++++ SHA HASH \
            ++++++++++++++++++++++++++++++++++\n'
         # <has value>'
                    <nxdl part>
    """
    sha_hash = ""
    with open(yaml_file, "r", encoding="utf-8") as inp_file:
        lines = inp_file.readlines()
        # file to write yaml part
        with open(sep_yaml, "w", encoding="utf-8") as yml_f_ob, open(
            sep_xml, "w", encoding="utf-8"
        ) as xml_f_ob:
            write_on_yaml = True

            last_line = lines[0]
            for line in lines[1:]:
                # Write in file when ensured that the next line is not with '++ SHA HASH ++'
                if "++ SHA HASH ++" not in line and write_on_yaml:
                    yml_f_ob.write(last_line)
                    last_line = line
                elif "++ SHA HASH ++" in line:
                    write_on_yaml = False
                    last_line = ""
                elif not write_on_yaml and not last_line:
                    # The first line of xml file has been found so in future write lines directly
                    # into xml file.
                    if not sha_hash:
                        sha_hash = line.split("# ", 1)[-1].strip()
                    else:
                        xml_f_ob.write(line[2:])
            # If the yaml fiile does not contain any hash for nxdl then we may have last line.
            if last_line:
                yml_f_ob.write(last_line)

    return sha_hash


def is_dom_comment(text):
    """Analyze a comment, whether it is a dom comment or not.

    Return true if dom comment.
    """

    # some signature keywords to distingush dom comments from other comments.
    signature_keyword_list = [
        "NeXus",
        "GNU Lesser General Public",
        "Free Software Foundation",
        "Copyright (C)",
        "WITHOUT ANY WARRANTY",
    ]
    for keyword in signature_keyword_list:
        if keyword not in text:
            return False

    return True



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
This file collects the functions for conversion from nxdl.xml to yaml version.
"""

# pylint: disable=too-many-lines

import re
import textwrap
from pathlib import Path
from typing import Callable, Dict, List

import lxml.etree as ET

from nyaml.helper import (
    NXDL_ATTRIBUTES_ATTRIBUTES,
    NXDL_FIELD_ATTRIBUTES,
    NXDL_GROUP_ATTRIBUTES,
    NXDL_LINK_ATTRIBUTES,
    clean_empty_lines,
    get_node_parent_info,
    get_yaml_escape_char_dict,
    is_dom_comment,
    remove_namespace_from_tag,
)

DEPTH_SIZE = 2 * " "
CMNT_TAG = "!--"
CMNT_TAG_END = "--"
CMNT_START = "<!--"
CMNT_END = "-->"
DEFINITION_CATEGORIES = ("category: application", "category: base")


def separate_pi_comments(input_file):
    """Separate PI comments from ProcessesInstruction (PI).

    ProcessesInstruction refers xml element process and version
    Separate the comments that comes immediately after XML process instruction part,
    i.e. copyright comment part.
    """
    comments_list = []
    comment = []

    with open(input_file, "r", encoding="utf-8") as file:
        lines = file.readlines()
        def_tag = "<definition"
        for line in lines:
            if CMNT_START in line:
                line = line.replace(CMNT_START, "")
                if CMNT_END in line:
                    line = line.replace(CMNT_END, "")
                    comments_list.append(line)
                else:
                    comment.append(line)
            elif CMNT_END in line and len(comment) > 0:
                comment.append(line.replace(CMNT_END, ""))
                comments_list.append("".join(comment))
                comment.clear()
            elif len(comment) > 0:
                comment.append(line)
            elif def_tag in line:
                break
    return comments_list


# Collected: https://dustinoprea.com/2019/01/22/python-parsing-xml-and-retaining-the-comments/
class _CommentedTreeBuilder(ET.TreeBuilder):
    def start(self, tag, attrs):
        super().start(tag=tag, attrs=attrs)

    def Comment(self, text):  # pylint: disable=invalid-name
        """Defining comment builder in TreeBuilder"""
        self.start(CMNT_TAG, {})
        self.data(text)
        self.end(CMNT_TAG_END)


def parse(filepath):
    """Parse xml function.

    Construct parser function for modified tree builder for including modified TreeBuilder
    and rebuilding XMLParser.
    """
    comments = separate_pi_comments(filepath)
    ctb = _CommentedTreeBuilder()
    xp_parser = ET.XMLParser(target=ctb, encoding="UTF-8")
    root = ET.parse(filepath, xp_parser)
    return comments, root


def handle_mapping_char(text, depth=-1, skip_n_line_on_top=False):
    """Check for escape character and replace by alternative character."""

    escape_char = get_yaml_escape_char_dict()
    for esc_key, val in escape_char.items():
        if esc_key in text:
            text = text.replace(esc_key, val)
    if not skip_n_line_on_top:
        if depth > 0:
            text = add_new_line_with_pipe_on_top(text, depth)
        else:
            raise ValueError("Need depth size to co-ordinate text line in yaml file.")
    return text


def add_new_line_with_pipe_on_top(text, depth):
    """Design docs block.

    Restructure the text by adding pipe '|' and '\n' char if ':' is found in text.
    """
    char = ":"
    if char in text:
        return "|" + "\n" + depth * DEPTH_SIZE + text
    return text


# pylint: disable=too-many-instance-attributes, too-many-public-methods
class Nxdl2yaml:
    """Parse XML file and print a YML file."""

    def __init__(
        self,
        symbol_list: List[str],
        root_level_definition: List[str],
        root_level_doc="",
        root_level_symbols="",
    ):
        # updated part of yaml_dict
        self.found_definition = False
        self.root_level_doc = root_level_doc
        self.root_level_symbols = root_level_symbols
        self.root_level_definition = root_level_definition
        self.symbol_list = symbol_list
        self.include_comment = True
        self.pi_comments = None
        # NOTE: Here is how root_level_comments organised for storing comments
        # root_level_comment= {'root_doc': comment,
        #                      'symbols': comment,
        #       The 'symbol_doc_comments' list is for comments from all 'symbol doc'
        #                      'symbol_doc_comments' : [comments]
        #                      'symbol_list': [symbols],
        #       The 'symbol_comments' contains comments for 'symbols doc' and all 'symbol'
        #                      'symbol_comments': [comments]}
        self.root_level_comment: Dict[str, str] = {}

        self.optionality_keys = (
            "minOccurs",
            "maxOccurs",
            "optional",
            "recommended",
            "required",
        )
        # "Take care of general attributes. Note other choices may be allowed in the future"
        self.choice_allowed_attr = ()

    def print_yml(self, input_file, output_yml, verbose):
        """Parse an XML file provided as input and print a YML file."""
        output_file_path = Path(output_yml)
        if output_file_path.is_file():
            output_file_path.unlink()

        depth = 0

        self.pi_comments, root = parse(input_file)
        xml_tree = {"tree": root, "node": root}
        self.xmlparse(output_yml, xml_tree, depth, verbose)

    def handle_symbols(self, depth, node):
        """Handle symbols field and its childs symbol"""

        self.root_level_symbols = (
            f"{remove_namespace_from_tag(node.tag)}: "
            f"{node.text.strip() if node.text else ''}"
        )
        depth += 1
        last_comment = ""
        sbl_doc_cmnt_list = []
        # Comments that come above symbol tag
        symbol_cmnt_list = []
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            if tag == CMNT_TAG and self.include_comment:
                last_comment = self.convert_to_yaml_comment(depth, child.text)
            if tag == "doc":
                symbol_cmnt_list.append(last_comment)
                # The line below is for handling length of 'symbol_comments' and
                # 'symbol_doc_comments'. Otherwise print_root_level_info() gets inconsistency
                # over for the loop while writing comment on file
                sbl_doc_cmnt_list.append("")
                last_comment = ""
                self.symbol_list.append(
                    self.handle_not_root_level_doc(depth, text=child.text)
                )
            elif tag == "symbol":
                # place holder is symbol name
                symbol_cmnt_list.append(last_comment)
                last_comment = ""
                if "doc" in child.attrib:
                    self.symbol_list.append(
                        self.handle_not_root_level_doc(
                            depth, tag=child.attrib["name"], text=child.attrib["doc"]
                        )
                    )
                else:
                    for symbol_doc in list(child):
                        tag = remove_namespace_from_tag(symbol_doc.tag)
                        if tag == CMNT_TAG and self.include_comment:
                            last_comment = self.convert_to_yaml_comment(
                                depth, symbol_doc.text
                            )
                        if tag == "doc":
                            sbl_doc_cmnt_list.append(last_comment)
                            last_comment = ""
                            self.symbol_list.append(
                                self.handle_not_root_level_doc(
                                    depth,
                                    tag=child.attrib["name"],
                                    text=symbol_doc.text,
                                )
                            )
        self.store_root_level_comments("symbol_doc_comments", sbl_doc_cmnt_list)
        self.store_root_level_comments("symbol_comments", symbol_cmnt_list)

    def store_root_level_comments(self, holder, comment):
        """Store yaml text or section line and the comments inteded for that lines or section"""

        self.root_level_comment[holder] = comment

    def handle_definition(self, node):
        """Handle definition group and its attributes.

        NOTE: Here we try to store the order of the xml element attributes, so that we get
        the same order in nxdl from yaml.
        """
        keyword = ""
        # tmp_word for reseving the location
        tmp_word = "#xx#"
        attribs = node.attrib
        # for tracking the order of name and type
        keyword_order = -1
        for item in attribs:
            if "name" == item:
                keyword = keyword + attribs[item]
                if keyword_order == -1:
                    self.root_level_definition.append(tmp_word)
                    keyword_order = self.root_level_definition.index(tmp_word)
            elif "extends" == item:
                keyword = f"{keyword}({attribs[item]})"
                if keyword_order == -1:
                    self.root_level_definition.append(tmp_word)
                    keyword_order = self.root_level_definition.index(tmp_word)
            elif "schemaLocation" not in item and "extends" != item:
                text = f"{item}: {attribs[item]}"
                self.root_level_definition.append(text)
        self.root_level_definition[keyword_order] = f"{keyword}:"

    def handle_root_level_doc(self, node):
        """
        Handle the documentation field found at root level.
        """
        text = self.handle_not_root_level_doc(depth=0, text=node.text)
        self.root_level_doc = text

    def clean_and_organise_text(self, text, depth):
        """Reconstruct text from doc and comment.

        Cleaninig up unintentional and accidental empty lines and spaces.
        """
        # Handling empty doc
        if not text:
            text = ""
        else:
            text = handle_mapping_char(text, -1, True)
        if "\n" in text:
            # To remove '\n' with non-space character as it will be added before text.
            text = clean_empty_lines(text.split("\n"))
            text_tmp = []
            yaml_indent_n = len((depth + 1) * DEPTH_SIZE)

            # Find indentation in the first line of text having alphabet
            first_line_indent_n = 0
            for line in text:
                # Consider only the lines that has at least one non-space character
                # and skip starting lines of a text block are empty
                if len(line.lstrip()) != 0:
                    first_line_indent_n = len(line) - len(line.lstrip())
                    break
            # Taking care of doc like below:
            # <doc>Text lines
            # text continues</doc>
            # So no indentation at the start of doc. So doc group will come along general
            # alignment
            if first_line_indent_n == 0:
                first_line_indent_n = yaml_indent_n

            # for indent_diff -ve all lines will move left by the same amount
            # for indent_diff +ve all lines will move right by the same amount
            indent_diff = yaml_indent_n - first_line_indent_n
            # CHeck for first line empty if not keep first line empty

            for line in text:
                line_indent_n = 0
                # count first empty spaces without alphabet
                line_indent_n = len(line) - len(line.lstrip())
                line_indent_n = line_indent_n + indent_diff
                if line_indent_n < yaml_indent_n:
                    # if line still under yaml identation
                    text_tmp.append(yaml_indent_n * " " + line.strip())
                else:
                    text_tmp.append(line_indent_n * " " + line.strip())

            text = "\n" + "\n".join(text_tmp)

        elif text:
            text = "\n" + (depth + 1) * DEPTH_SIZE + text.strip()

        return text

    def check_and_handle_doc_xref_and_other_doc(self, text, indent):
        """Check for xref doc which comes as a block of text.

        The doc part below is the example how xref comes:
        '''
        This concept is related to term `<term>`_ of the <spec> standard.
        .. _<term>: <url>
        '''
        converter as
        '''
        <indent>  "xref:
        <indent>    xpec: <value>
        <indent>    erm: <value>
        <indent>    url: <value>"
        '''

        Parameters
        ----------
        text: str
            plain text.
        Returns
        -------
        str
            return part of doc as formatted
        """

        xref_key, spec_key, term_key, url_key = ("xref", "spec", "term", "url")
        spec, term, url = (None, None, None)
        matches = re.search(
            r"This concept is related to term `([^`:]+)`_ of the"
            r" (.*?) standard\.\s+\.\. _\1: ([^\s]+)",
            text,
        )
        if matches:
            term = matches.group(1)
            spec = matches.group(2)
            url = matches.group(3)
            indent = indent + DEPTH_SIZE  # see example in func doc
            return (
                f'{indent}"{xref_key}:\n{indent + DEPTH_SIZE}{spec_key}: {spec}'
                f"\n{indent + DEPTH_SIZE}{term_key}"
                f': {term}\n{indent + DEPTH_SIZE}{url_key}: {url}"'
            ), True
        return text, False

    # pylint: disable=too-many-branches, too-many-locals
    def handle_not_root_level_doc(self, depth, text, tag="doc", file_out=None):
        """Handle docs field of group and field but not root.

        Handle docs field along the yaml file. In this function we also tried to keep
        the track of indentation. E.g. the below doc block.
            * Topic name
                Description of topic
        """
        if "}" in tag:
            tag = remove_namespace_from_tag(tag)
        indent = depth * DEPTH_SIZE
        text = self.clean_and_organise_text(text, depth)  # starts with '\n'
        docs = re.split(r"\n\s*\n", text)
        parts = []

        # Add links to previous docstring
        for i, doc in enumerate(docs):
            link_match = re.match(r"\s*\.\. _.*", doc)
            if link_match is not None:
                parts[i - 1] += doc
            else:
                parts.append(doc)

        modified_docs = []
        xref_in_doc = False
        for doc_part in parts:
            if not doc_part.isspace():
                mod_doc, xref_present = self.check_and_handle_doc_xref_and_other_doc(
                    doc_part, indent
                )
                xref_in_doc = xref_in_doc or xref_present
                modified_docs.append(mod_doc)
        # doc example:
        # doc:
        #  - |
        #   text
        #  - |
        #   xref:
        #       spec:
        #       term:
        if len(modified_docs) == 1:
            doc_str = f"{indent}{tag}: |{modified_docs[0]}\n"
        elif len(modified_docs) > 1 and xref_in_doc:
            doc_str = f"{indent}{tag}:\n"
            for mod_doc in modified_docs:
                if not re.match(
                    r"^\s*\n", mod_doc
                ):  # if not starts with 'spaces and/or \n'
                    mod_doc = "\n" + mod_doc
                # doc_str = f"{doc_str}{indent} - |\n{textwrap.indent(mod_doc, indent+'  ')}\n"
                doc_str = f"{doc_str}{indent} - |{textwrap.indent(mod_doc, '')}\n"
        else:
            doc_str = f"{indent}{tag}: |{text}\n"

        if file_out:
            file_out.write(doc_str)
            return None
        return doc_str

    def write_out(self, indent, text, file_out):
        """
        Write text line in output file.
        """
        line_string = f"{indent}{text.rstrip()}\n"
        file_out.write(line_string)

    def print_root_level_doc(self, file_out):
        """Print at the root level of YML file the general documentation field found in XML file"""
        indent = 0 * DEPTH_SIZE

        text = self.root_level_comment.get("root_doc")
        if text:
            self.write_out(indent, text, file_out)

        text = self.root_level_doc
        self.write_out(indent, text, file_out)
        self.root_level_doc = ""

    def convert_to_yaml_comment(self, depth, text):
        """
        Convert into yaml comment by adding exta '#' char in front of comment lines
        """
        # To store indentation text from comment
        lines = self.clean_and_organise_text(text, depth).split("\n")
        indent = DEPTH_SIZE * depth
        mod_lines = []
        for line in lines:
            line = line.strip()
            if line:
                if line[0] != "#":
                    line = "# " + line
                mod_lines.append(line)
        # The starting '\n' to keep multiple comments separate
        return "\n" + "\n".join(map(lambda line: indent + line, mod_lines))

    def print_root_level_info(self, depth, file_out):
        """
        Print at the root level of YML file \
        the information stored as definition attributes in the XML file
        """
        if depth < 0:
            raise ValueError("Somthing wrong with indentation in root level.")

        has_category = False
        for def_line in self.root_level_definition:
            if def_line in DEFINITION_CATEGORIES:
                self.write_out(indent=0 * DEPTH_SIZE, text=def_line, file_out=file_out)
                has_category = True

        if not has_category:
            raise ValueError(
                "Definition dose not get any category from 'base or application'."
            )
        self.print_root_level_doc(file_out)
        text = self.root_level_comment.get("symbols", "")
        if text:
            indent = depth * DEPTH_SIZE
            self.write_out(indent, text, file_out)
        if self.root_level_symbols:
            self.write_out(
                indent=0 * DEPTH_SIZE, text=self.root_level_symbols, file_out=file_out
            )
            # symbol_list include 'symbols doc', and all 'symbol'
            for ind, symbol in enumerate(self.symbol_list):
                # Taking care of comments that come on top of 'symbols doc' and 'symbol'
                if (
                    "symbol_comments" in self.root_level_comment
                    and self.root_level_comment["symbol_comments"][ind] != ""
                ):
                    indent = depth * DEPTH_SIZE
                    self.write_out(
                        indent,
                        self.root_level_comment["symbol_comments"][ind],
                        file_out,
                    )
                if (
                    "symbol_doc_comments" in self.root_level_comment
                    and self.root_level_comment["symbol_doc_comments"][ind] != ""
                ):
                    indent = depth * DEPTH_SIZE
                    self.write_out(
                        indent,
                        self.root_level_comment["symbol_doc_comments"][ind],
                        file_out,
                    )

                self.write_out(indent=(0 * DEPTH_SIZE), text=symbol, file_out=file_out)
        indent = depth * DEPTH_SIZE
        # Take care copyright doc string
        for comment in self.pi_comments:
            if comment and not is_dom_comment(comment):
                self.write_out(
                    indent, self.convert_to_yaml_comment(depth, comment), file_out
                )
        if self.root_level_definition:
            # Store NXname for writing at end of definition attributes
            nx_name = ""
            for defs in self.root_level_definition:
                if "NX" in defs and defs[-1] == ":":
                    nx_name = defs
                    continue
                if defs in ("category: application", "category: base"):
                    continue
                self.write_out(indent=0 * DEPTH_SIZE, text=defs, file_out=file_out)
            self.write_out(indent=0 * DEPTH_SIZE, text=nx_name, file_out=file_out)
        self.found_definition = False

    def handle_exists(self, exists_dict, key, val):
        """
        Create exist component as folows:

        {'min' : value for min,
         'max' : value for max,
         'optional' : value for optional}

        This is created separately so that the keys stays in order.
        """
        if not val:
            val = ""
        else:
            val = str(val)
        if "minOccurs" == key:
            exists_dict["minOccurs"] = ["min", val]
        elif "maxOccurs" == key:
            exists_dict["maxOccurs"] = ["max", val]
        elif "optional" == key:
            exists_dict["optional"] = ["optional", val]
        elif "recommended" == key:
            exists_dict["recommended"] = ["recommended", val]
        elif "required" == key:
            exists_dict["required"] = ["required", val]

    # pylint: disable=too-many-branches
    def handle_group_or_field(self, depth, node, file_out):
        """Handle all the possible attributes that come along a field or group"""

        name_type = ""
        node_attr = node.attrib
        rm_key_list = []
        # Maintain order: name and type in form name(type) or (type)name that come first
        for key, val in node_attr.items():
            if key == "name":
                name_type = name_type + val
                rm_key_list.append(key)
            elif key == "type":
                name_type = f"{name_type}({val})"
                rm_key_list.append(key)
        if not name_type:
            raise ValueError(
                f"No 'name' or 'type' has been found. But, 'group' or 'field' "
                f"must have at least a name.We have attributes:  {node_attr}"
            )
        indent = depth * DEPTH_SIZE
        file_out.write(f"{indent}{name_type}:\n")

        for key in rm_key_list:
            del node_attr[key]

        # tmp_dict intended to preserve order of attributes
        tmp_dict = {}
        exists_dict = {}
        for key, val in node_attr.items():
            # Check for any unwanted attributes
            self.check_for_unwanted_attributes(node=node)
            # As both 'minOccurs', 'maxOccurs' and optionality move to the 'exists'
            if key in self.optionality_keys:
                if "exists" not in tmp_dict:
                    tmp_dict["exists"] = []
                self.handle_exists(exists_dict, key, val)
            elif key == "units":
                tmp_dict["unit"] = str(val)
            else:
                tmp_dict[key] = str(val)

        if exists_dict:
            for key, val in exists_dict.items():
                if key in ["minOccurs", "maxOccurs"]:
                    tmp_dict["exists"] = tmp_dict["exists"] + val
                elif key in ["optional", "recommended", "required"]:
                    tmp_dict["exists"] = key

        depth_ = depth + 1
        for key, val in tmp_dict.items():
            # Increase depth size inside handle_map...() for writing text with one
            # more indentation.
            file_out.write(
                f"{depth_ * DEPTH_SIZE}{key}: "
                f"{handle_mapping_char(val, depth_ + 1, False)}\n"
            )

    def check_for_unwanted_attributes(self, node, allowed_attributes_li=None, tag=None):
        """Check for any attributes that NeXus does not allow."""
        node_tag = remove_namespace_from_tag(node.tag)
        if node_tag == "field":
            for key in node.attrib.keys():
                if key not in NXDL_FIELD_ATTRIBUTES:
                    raise ValueError(
                        f"Field has an unwanted attribute {key}."
                        f"NeXus field allows attributes from {NXDL_FIELD_ATTRIBUTES}"
                    )
        elif node_tag == "group":
            for key in node.attrib.keys():
                if key not in NXDL_GROUP_ATTRIBUTES:
                    raise ValueError(
                        f"Attribute has an unwanted attribute {key}."
                        f"NeXus attribute allows attributes from {NXDL_GROUP_ATTRIBUTES}"
                    )
        elif node_tag == tag:
            for key in node.attrib.keys():
                if key not in allowed_attributes_li:
                    raise ValueError(
                        f"{tag.capitalized()} has an unwanted attribute {key}."
                        f"NeXus {tag.capitalized()} allows attributes from {allowed_attributes_li}"
                    )

    # pylint: disable=too-many-branches, too-many-locals, too-many-statements
    def handle_dimension(self, depth, node, file_out):
        """Handle the dimension field.

            NOTE: Usually we take care of any xml element in xmlparse(...) and
        recursion_in_xml_tree(...) functions. But here it is a bit different: the doc dimension
          and attributes of dim has been handled inside this function here.
        """
        possible_dim_attrs = ["ref", "required", "incr", "refindex"]
        possible_dimemsion_attrs = ["rank"]

        # taking care of Dimension tag
        indent = depth * DEPTH_SIZE
        tag = remove_namespace_from_tag(node.tag)
        file_out.write(f"{indent}{tag}:\n")
        for attr, value in node.attrib.items():
            if attr in possible_dimemsion_attrs and not isinstance(value, dict):
                indent = (depth + 1) * DEPTH_SIZE
                file_out.write(f"{indent}{attr}: {value}\n")
            else:
                raise ValueError(
                    f"Dimension has an attribute {attr} that is not valid."
                    f"Current the allowd atributes are {possible_dimemsion_attrs}."
                    f" Please have a look"
                )
        # taking care of dimension doc
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            if tag == "doc":
                text = self.handle_not_root_level_doc(depth + 1, child.text)
                file_out.write(text)
                node.remove(child)

        dim_index_value = ""
        dim_other_parts = {}
        dim_cmnt_node = []
        # taking care of dim and doc children of dimension
        for child in list(node):
            tag = remove_namespace_from_tag(child.tag)
            child_attrs = child.attrib
            # taking care of index and value attributes
            if tag == "dim":
                # taking care of index and value in format [[index, value]]
                index = child_attrs.pop("index", "")
                value = child_attrs.pop("value", "")
                dim_index_value = f"{dim_index_value}[{index}, {value}], "

                # Taking care of doc comes as child of dim
                for cchild in list(child):
                    ttag = cchild.tag.split("}", 1)[1]
                    if ttag == ("doc"):
                        if ttag not in dim_other_parts:
                            dim_other_parts[ttag] = []
                        text = cchild.text
                        dim_other_parts[ttag].append(text.strip())
                        child.remove(cchild)
                        continue
                # taking care of other attributes except index and value
                for attr, value in child_attrs.items():
                    if attr in possible_dim_attrs:
                        if attr not in dim_other_parts:
                            dim_other_parts[attr] = []
                        dim_other_parts[attr].append(value)
            if tag == CMNT_TAG and self.include_comment:
                # Store and remove node so that comment nodes from dim node so
                # that it does not call in xmlparser function
                dim_cmnt_node.append(child)
                node.remove(child)

        # All 'dim' element comments on top of 'dim' yaml key
        if dim_cmnt_node:
            for ch_nd in dim_cmnt_node:
                self.handle_comment(depth + 1, ch_nd, file_out)
        # index and value attributes of dim elements
        indent = (depth + 1) * DEPTH_SIZE
        value = dim_index_value[:-2] or ""
        file_out.write(f"{indent}dim: [{value}]\n")

        # Write the attributes, except index and value, and doc of dim as child of dim_parameter.
        # But the doc or attributes for each dim come inside list according to the order of dim.
        if dim_other_parts:
            indent = (depth + 1) * DEPTH_SIZE
            file_out.write(f"{indent}dim_parameters:\n")
            # depth = depth + 2 dim_parameter has child such as doc of dim
            indent = (depth + 2) * DEPTH_SIZE
            for key, value in dim_other_parts.items():
                if key == "doc":
                    value = self.handle_not_root_level_doc(
                        depth + 2, str(value), key, file_out
                    )
                else:
                    # Increase depth size inside handle_map...() for writing text with one
                    # more indentation.
                    file_out.write(
                        f"{indent}{key}: "
                        f"{handle_mapping_char(value, depth + 3, False)}\n"
                    )

    def handle_enumeration(self, depth, node, file_out):
        """
            Handle the enumeration field parsed from the xml file.

        If the enumeration items contain a doc field, the yaml file will contain items as child
        fields of the enumeration field.

        If no doc are inherited in the enumeration items, a list of the items is given for the
        enumeration list.

        """

        check_doc = False
        node_children = list(node)
        for child in node_children:
            if list(child):
                check_doc = True
                break
        # pylint: disable=too-many-nested-blocks
        if check_doc:
            indent = depth * DEPTH_SIZE
            tag = remove_namespace_from_tag(node.tag)
            file_out.write(f"{indent}{tag}: \n")
            for child in node_children:
                tag = remove_namespace_from_tag(child.tag)
                itm_depth = depth + 1
                if tag == "item":
                    indent = itm_depth * DEPTH_SIZE
                    value = child.attrib["value"]
                    file_out.write(f"{indent}{value}: \n")
                    child_children = list(child)
                    if child_children:
                        for item_doc in child_children:
                            if remove_namespace_from_tag(item_doc.tag) == "doc":
                                item_doc_depth = itm_depth + 1
                                self.handle_not_root_level_doc(
                                    item_doc_depth,
                                    item_doc.text,
                                    item_doc.tag,
                                    file_out,
                                )
                            if (
                                remove_namespace_from_tag(item_doc.tag) == CMNT_TAG
                                and self.include_comment
                            ):
                                self.handle_comment(itm_depth + 1, item_doc, file_out)
                if tag == CMNT_TAG and self.include_comment:
                    self.handle_comment(itm_depth + 1, child, file_out)
        else:
            enum_list = []
            remove_nodes = []
            for item_child in node_children:
                tag = remove_namespace_from_tag(item_child.tag)
                if tag == "item":
                    value = item_child.attrib["value"]
                    enum_list.append(value)
                if tag == CMNT_TAG and self.include_comment:
                    self.handle_comment(depth, item_child, file_out)
                    remove_nodes.append(item_child)
            for ch_node in remove_nodes:
                node.remove(ch_node)

            indent = depth * DEPTH_SIZE
            tag = remove_namespace_from_tag(node.tag)
            enum_list = ", ".join(enum_list)
            file_out.write(f"{indent}{tag}: [{enum_list}]\n")

    def handle_attributes(self, depth, node, file_out):
        """Handle the attributes parsed from the xml file"""

        name = ""
        nm_attr = "name"
        node_attr = node.attrib
        # Maintain order: name and type in form name(type) or (type)name that come first
        name = node_attr.pop(nm_attr, "")
        if not name:
            raise ValueError("Attribute must have an name key.")

        indent = depth * DEPTH_SIZE
        escapesymbol = r"\@"
        file_out.write(f"{indent}{escapesymbol}{name}:\n")

        tmp_dict = {}
        exists_dict = {}
        for key, val in node_attr.items():
            if key not in NXDL_ATTRIBUTES_ATTRIBUTES:
                raise ValueError(
                    f"An attribute ({key}) has been found that is not allowed."
                    f"The allowed attr is {NXDL_ATTRIBUTES_ATTRIBUTES}."
                )
            # As both 'minOccurs', 'maxOccurs' and optionality move to the 'exists'
            if key in self.optionality_keys:
                if "exists" not in tmp_dict:
                    tmp_dict["exists"] = []
                self.handle_exists(exists_dict, key, val)
            elif key == "units":
                tmp_dict["unit"] = val
            else:
                tmp_dict[key] = val

        has_min_max = False
        has_opt_reco_requ = False
        if exists_dict:
            for key, val in exists_dict.items():
                if key in ["minOccurs", "maxOccurs"]:
                    tmp_dict["exists"] = tmp_dict["exists"] + val
                    has_min_max = True
                elif key in ["optional", "recommended", "required"]:
                    tmp_dict["exists"] = key
                    has_opt_reco_requ = True
        if has_min_max and has_opt_reco_requ:
            raise ValueError(
                "Optionality 'exists' can take only either from ['minOccurs',"
                " 'maxOccurs'] or from ['optional', 'recommended', 'required']"
                ". But not from both of the groups together. Please check in"
                " attributes"
            )

        depth_ = depth + 1
        for key, val in tmp_dict.items():
            # Increase depth size inside handle_map...() for writing text with one
            # more indentation.
            file_out.write(
                f"{depth_ * DEPTH_SIZE}{key}: "
                f"{handle_mapping_char(val, depth_ + 1, False)}\n"
            )

    def handle_link(self, depth, node, file_out):
        """Handle link elements of nxdl"""

        node_attr = node.attrib
        # Handle special cases
        name = node_attr.pop("name", "")
        if name:
            indent = depth * DEPTH_SIZE
            file_out.write(f"{indent}{name}(link):\n")

        depth_ = depth + 1
        # Handle general cases
        for attr_key, val in node_attr.items():
            if attr_key in NXDL_LINK_ATTRIBUTES:
                indent = depth_ * DEPTH_SIZE
                file_out.write(f"{indent}{attr_key}: {val}\n")
            else:
                raise ValueError(
                    f"An unexpected attribute '{attr_key}' of link has found."
                    f"At this moment the allowed keys are {NXDL_LINK_ATTRIBUTES}"
                )

    def handle_choice(self, depth, node, file_out):
        """
        Handle choice element which is a parent node of group.
        """

        node_attr = node.attrib
        name = node_attr.pop("name", "")
        # Handle special casees
        if name:
            indent = depth * DEPTH_SIZE
            file_out.write(f"{indent}{name}(choice): \n")

        depth_ = depth + 1
        # Take care of attributes of choice element, attributes may come in future.
        for attr in node_attr.items():
            if attr in self.choice_allowed_attr:
                indent = depth_ * DEPTH_SIZE
                value = node_attr[attr]
                file_out.write(f"{indent}{attr}: {value}\n")
            else:
                raise ValueError(
                    f"An unexpected attribute '{attr}' of 'choice' has been found."
                    f"At this moment allowed attributes for choice {self.choice_allowed_attr}"
                )

    def handle_comment(self, depth, node, file_out):
        """
        Collect comment element and pass to write_out function
        """
        indent = depth * DEPTH_SIZE
        text = self.convert_to_yaml_comment(depth, node.text)
        self.write_out(indent, text, file_out)

    def recursion_in_xml_tree(self, depth, xml_tree, output_yml, verbose):
        """
            Descend lower level in xml tree. If we are in the symbols branch, the recursive
        behaviour is not triggered as we already handled the symbols' children.
        """

        tree = xml_tree["tree"]
        node = xml_tree["node"]
        for child in list(node):
            xml_tree_children = {"tree": tree, "node": child}
            self.xmlparse(output_yml, xml_tree_children, depth, verbose)

    # pylint: disable=too-many-branches, too-many-statements
    def xmlparse(self, output_yml, xml_tree, depth, verbose):
        """
        Main method of the nxdl2yaml converter.
        It parses XML tree, then prints recursively each level of the tree
        """
        tree = xml_tree["tree"]
        node = xml_tree["node"]
        if verbose:
            if isinstance(node.tag, Callable):
                print(f"Node tag: {node.tag}\n")
                print(f"Node text: {node.text}\n")
            else:
                print(f"Node tag: {remove_namespace_from_tag(node.tag)}\n")
                print(f"Attributes: {node.attrib}\n")
        with open(output_yml, "a", encoding="utf-8") as file_out:
            tag = remove_namespace_from_tag(node.tag)
            if tag == "definition":
                self.found_definition = True
                self.handle_definition(node)
                # Taking care of root level doc and symbols
                remove_cmnt_n = None
                last_comment = ""
                for child in node:
                    tag_tmp = remove_namespace_from_tag(child.tag)
                    if tag_tmp == CMNT_TAG and self.include_comment:
                        last_comment = self.convert_to_yaml_comment(depth, child.text)
                        remove_cmnt_n = child
                    if tag_tmp == "doc":
                        self.store_root_level_comments("root_doc", last_comment)
                        last_comment = ""
                        self.handle_root_level_doc(child)
                        node.remove(child)
                        if remove_cmnt_n is not None:
                            node.remove(remove_cmnt_n)
                            remove_cmnt_n = None
                    if tag_tmp == "symbols":
                        self.store_root_level_comments("symbols", last_comment)
                        last_comment = ""
                        self.handle_symbols(depth, child)
                        node.remove(child)
                        if remove_cmnt_n is not None:
                            node.remove(remove_cmnt_n)
                            remove_cmnt_n = None

            if tag == "doc" and depth != 1:
                parent = get_node_parent_info(tree, node)[0]
                doc_parent = remove_namespace_from_tag(parent.tag)
                if doc_parent != "item":
                    self.handle_not_root_level_doc(
                        depth, text=node.text, tag=node.tag, file_out=file_out
                    )

            if self.found_definition is True and self.root_level_doc:
                self.print_root_level_info(depth, file_out)
            # End of print root-level definitions in file
            if tag in ("field", "group") and depth != 0:
                self.handle_group_or_field(depth, node, file_out)
            if tag == ("enumeration"):
                self.handle_enumeration(depth, node, file_out)
            if tag == ("attribute"):
                self.handle_attributes(depth, node, file_out)
            if tag == ("dimensions"):
                self.handle_dimension(depth, node, file_out)
            if tag == ("link"):
                self.handle_link(depth, node, file_out)
            if tag == ("choice"):
                self.handle_choice(depth, node, file_out)
            if tag == CMNT_TAG and self.include_comment:
                self.handle_comment(depth, node, file_out)
        depth += 1
        # Write nested nodes
        self.recursion_in_xml_tree(depth, xml_tree, output_yml, verbose)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""
Tests for nyaml2nxdl tool
"""
import filecmp
import os
import sys
from datetime import datetime
from pathlib import Path

import lxml.etree as ET
import pytest
from click.testing import CliRunner

from nyaml import cli as nyaml2nxdl
from nyaml import nyaml2nxdl as nyaml2nxdl_forward_tools
from nyaml.comment_collector import CommentCollector
from nyaml.helper import LineLoader, remove_namespace_from_tag
from nyaml.nyaml2nxdl import handle_each_part_doc


def delete_duplicates(list_of_matching_string):
    """
    Delete duplicate from lists
    """
    return list(dict.fromkeys(list_of_matching_string))


def check_file_fresh_baked(test_file):
    """
    Get sure that the test file is generated by the converter
    """
    path = Path(test_file)
    timestamp = datetime.fromtimestamp(path.stat().st_mtime).strftime("%d/%m/%Y %H:%M")
    now = datetime.now().strftime("%d/%m/%Y %H:%M")
    assert timestamp == now, "xml file not generated"


def find_matches(xml_file, desired_matches):
    """
        Read xml file and find desired matches. Return a list of two lists in the form:
    [[matching_line],[matching_line_index]]
    """
    with open(xml_file, "r", encoding="utf-8") as file:
        xml_reference = file.readlines()
    lines = []
    lines_index = []
    found_matches = []
    for i, line in enumerate(xml_reference):
        for desired_match in desired_matches:
            if str(desired_match) in str(line):
                lines.append(line)
                lines_index.append(i)
                found_matches.append(desired_match)
    # ascertain that all the desired matches were found in file
    found_matches_clean = delete_duplicates(found_matches)
    assert len(found_matches_clean) == len(
        desired_matches
    ), "some desired_matches were \nnot found in file"
    return [lines, lines_index]


def compare_matches(ref_xml_file, test_yml_file, test_xml_file, desired_matches):
    """
        Check if a new xml file is generated
    and if test xml file is equal to reference xml file
    """
    # Reference file is read
    ref_matches = find_matches(ref_xml_file, desired_matches)
    # Test file is generated
    runner = CliRunner()
    result = runner.invoke(nyaml2nxdl.launch_tool, [test_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)
    # Test file is read
    test_matches = find_matches(test_xml_file, desired_matches)
    assert test_matches == ref_matches


def test_links():
    """
    Check the correct parsing of links
    """
    ref_xml_link_file = "tests/data/Ref_NXtest_links.nxdl.xml"
    test_yml_link_file = "tests/data/NXtest_links.yaml"
    test_xml_link_file = "tests/data/NXtest_links.nxdl.xml"
    # ref_xml_link_file = os.path.abspath(data_path + '/Ref_NXtest_links.nxdl.xml')
    # test_yml_link_file = os.path.abspath(data_path + '/NXtest_links.yaml')
    # test_xml_link_file = os.path.abspath(data_path + '/NXtest_links.nxdl.xml')
    desired_matches = ["<link", "/>"]
    compare_matches(
        ref_xml_link_file, test_yml_link_file, test_xml_link_file, desired_matches
    )
    os.remove("tests/data/NXtest_links.nxdl.xml")
    sys.stdout.write("Test on links okay.\n")


def test_docs():
    """In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_xml_file = "tests/data/Ref_NXellipsometry-docCheck.nxdl.xml"
    test_yml_file = "tests/data/NXellipsometry-docCheck.yaml"
    test_xml_file = "tests/data/NXellipsometry-docCheck.nxdl.xml"
    desired_matches = ["<doc", "</doc>"]
    compare_matches(ref_xml_file, test_yml_file, test_xml_file, desired_matches)
    os.remove("tests/data/NXellipsometry-docCheck.nxdl.xml")
    sys.stdout.write("Test on documentation formatting okay.\n")


def test_nxdl2yaml_doc_format_and_nxdl_part_as_comment():
    """
    This test for two reason:
        1. In test-1 an nxdl file with all kind of doc formats are translated
    to yaml to check if they are correct.
        2. In test-2: Check the nxdl that comes at the end of yaml file as comment.
    """
    ref_xml_file = "tests/data/Ref_NXentry.nxdl.xml"
    ref_yml_file = "tests/data/Ref_NXentry.yaml"
    test_yml_file = "tests/data/Ref_NXentry_parsed.yaml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [ref_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)

    result = filecmp.cmp(ref_yml_file, test_yml_file, shallow=False)
    assert (
        result
    ), "Ref YML and parsed YML\
has not the same structure!!"
    os.remove(test_yml_file)
    sys.stdout.write("Test on xml -> yml doc formatting okay.\n")


def test_fileline_error():
    """
    In this test the yaml fileline in the error message is tested.
    """
    test_yml_file = "tests/data/NXfilelineError1.yaml"
    out_nxdl = "tests/data/NXfilelineError1.nxdl.xml"
    out_yaml = "tests/data/temp_NXfilelineError1.yaml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [test_yml_file])
    assert result.exit_code == 1
    assert "13" in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    test_yml_file = "tests/data/NXfilelineError2.yaml"
    out_nxdl = "tests/data/NXfilelineError2.nxdl.xml"
    out_yaml = "tests/data/temp_NXfilelineError2.yaml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [test_yml_file])
    assert result.exit_code == 1
    assert "21" in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    test_yml_file = "tests/data/NXfilelineError3.yaml"
    out_nxdl = "tests/data/NXfilelineError3.nxdl.xml"
    out_yaml = "tests/data/temp_NXfilelineError3.yaml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [test_yml_file])
    assert result.exit_code == 1
    assert "25" in str(result.exception)
    os.remove(out_nxdl)
    os.remove(out_yaml)

    sys.stdout.write("Test on xml -> yml fileline error handling okay.\n")


def test_symbols():
    """
    Check the correct parsing of symbols
    """
    ref_xml_symbol_file = "tests/data/Ref_NXnested_symbols.nxdl.xml"
    test_yml_symbol_file = "tests/data/NXnested_symbols.yaml"
    test_xml_symbol_file = "tests/data/NXnested_symbols.nxdl.xml"
    desired_matches = ["<symbols>", "</symbols>", "<symbols"]
    compare_matches(
        ref_xml_symbol_file, test_yml_symbol_file, test_xml_symbol_file, desired_matches
    )
    os.remove("tests/data/NXnested_symbols.nxdl.xml")
    sys.stdout.write("Test on symbols okay.\n")


def test_attributes():
    """
    Check expected attributes in NeXus fields, groups, and attributes.
    Check proper doc elements.
    """
    ref_xml_attribute_file = "tests/data/Ref_NXattributes.nxdl.xml"
    test_yml_attribute_file = "tests/data/NXattributes.yaml"
    test_xml_attribute_file = "tests/data/NXattributes.nxdl.xml"
    desired_matches = [
        "<attribute",
        "</attribute>",
        "<doc>",
        "</doc>",
        "<field",
        "</field>",
        "<group",
        "</group>",
    ]
    compare_matches(
        ref_xml_attribute_file,
        test_yml_attribute_file,
        test_xml_attribute_file,
        desired_matches,
    )
    os.remove("tests/data/NXattributes.nxdl.xml")
    sys.stdout.write("Test on attributes okay.\n")


def test_extends():
    """
    Check the correct handling of extends keyword
    """
    ref_xml_attribute_file = "tests/data/Ref_NXattributes.nxdl.xml"
    test_yml_attribute_file = "tests/data/NXattributes.yaml"
    test_xml_attribute_file = "tests/data/NXattributes.nxdl.xml"
    runner = CliRunner()
    result = runner.invoke(nyaml2nxdl.launch_tool, [test_yml_attribute_file])
    assert result.exit_code == 0
    ref_root_node = ET.parse(ref_xml_attribute_file).getroot()
    test_root_node = ET.parse(test_xml_attribute_file).getroot()
    assert ref_root_node.attrib == test_root_node.attrib
    os.remove("tests/data/NXattributes.nxdl.xml")
    sys.stdout.write("Test on extends keyword okay.\n")


def test_symbols_and_enum_docs():
    """
        Check the correct handling of empty attributes
    or attributes fields, e.g. doc
    """
    ref_xml_file = "tests/data/Ref_NXmytests.nxdl.xml"
    test_yml_file = "tests/data/NXmytests.yaml"
    test_xml_file = "tests/data/NXmytests.nxdl.xml"
    desired_matches = [
        "<attribute",
        "</attribute>",
        "<doc>",
        "</doc>",
        "<symbols>",
        "</symbols>",
        "<symbols",
        "<dimensions",
        "</dimensions>",
        "<dim",
    ]
    compare_matches(ref_xml_file, test_yml_file, test_xml_file, desired_matches)
    os.remove("tests/data/NXmytests.nxdl.xml")
    sys.stdout.write("Test on docs in enumeration and symbols okay.\n")


def test_xml_parsing():
    """
        In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_xml_file = "tests/data/Ref_NXellips.nxdl.xml"
    test_yml_file = "tests/data/Ref_NXellips_parsed.yaml"
    test_xml_file = "tests/data/Ref_NXellips_parsed.nxdl.xml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [ref_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [test_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)

    test_tree = ET.parse(test_xml_file)
    test_tree_flattened = {i.tag.split("}", 1)[1] for i in test_tree.iter()}

    ref_tree = ET.parse(ref_xml_file)
    ref_tree_flattened = {i.tag.split("}", 1)[1] for i in ref_tree.iter()}

    assert (
        test_tree_flattened == ref_tree_flattened
    ), "Ref XML and parsed XML\
has not the same tree structure!!"
    os.remove("tests/data/Ref_NXellips_parsed.nxdl.xml")
    os.remove("tests/data/Ref_NXellips_parsed.yaml")
    sys.stdout.write("Test on xml -> yml -> xml okay.\n")


def test_yml_parsing():
    """In this test an xml file in converted to yml and then back to xml.
    The xml trees of the two files are then compared.
    """
    ref_yml_file = "tests/data/Ref_NXellipsometry.yaml"
    test_xml_file = "tests/data/Ref_NXellipsometry.nxdl.xml"
    test_yml_file = "tests/data/Ref_NXellipsometry_parsed.yaml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [ref_yml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_xml_file)
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [test_xml_file])
    assert result.exit_code == 0
    check_file_fresh_baked(test_yml_file)

    test_yml_tree = nyaml2nxdl_forward_tools.yml_reader(test_yml_file)

    ref_yml_tree = nyaml2nxdl_forward_tools.yml_reader(ref_yml_file)

    assert list(test_yml_tree) == list(
        ref_yml_tree
    ), "Ref YML and parsed YML \
has not the same root entries!!"
    os.remove("tests/data/Ref_NXellipsometry_parsed.yaml")
    os.remove("tests/data/Ref_NXellipsometry.nxdl.xml")
    sys.stdout.write("Test on yml -> xml -> yml okay.\n")


def test_yml_consistency_comment_parsing():
    """Test comments parsing from yaml. Convert 'yaml' input file to '.nxdl.xml' and
    '.nxdl.xml' to '.yaml'
    """

    ref_yml_file = "tests/data/Ref_NXcomment.yaml"
    test_yml_file = "tests/data/Ref_NXcomment_consistency.yaml"

    result = CliRunner().invoke(
        nyaml2nxdl.launch_tool, ["--check-consistency", ref_yml_file]
    )
    assert result.exit_code == 0, (
        f"Exception: {result.exception}, \nExecution Info:" "{result.exc_info}"
    )
    with open(ref_yml_file, "r", encoding="utf-8") as ref_yml:
        loader = LineLoader(ref_yml)
        ref_loaded_yaml = loader.get_single_data()
    ref_comment_blocks = CommentCollector(ref_yml_file, ref_loaded_yaml)
    ref_comment_blocks.extract_all_comment_blocks()

    with open(test_yml_file, "r", encoding="utf-8") as test_yml:
        loader = LineLoader(test_yml)
        test_loaded_yaml = loader.get_single_data()
    test_comment_blocks = CommentCollector(test_yml_file, test_loaded_yaml)
    test_comment_blocks.extract_all_comment_blocks()

    for ref_cmnt, test_cmnt in zip(ref_comment_blocks, test_comment_blocks):
        assert ref_cmnt == test_cmnt, "Comment is not consistent."

    os.remove(test_yml_file)


def test_yml2xml_comment_parsing():
    """To test comment that written in xml for element attributes, e.g.
    attribute 'rank' for 'dimension' element and attribute 'exists' for
    'NXentry' group element.
    """
    input_yml = "tests/data/NXcomment_yaml2nxdl.yaml"
    ref_xml = "tests/data/Ref_NXcomment_yaml2nxdl.nxdl.xml"
    test_xml = "tests/data/NXcomment_yaml2nxdl.nxdl.xml"

    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [input_yml])
    assert result.exit_code == 0

    ref_root = ET.parse(ref_xml).getroot()
    test_root = ET.parse(test_xml).getroot()

    def recursive_compare(ref_root, test_root):
        assert ref_root.attrib.items() == test_root.attrib.items(), (
            "Got different xml element" "Atribute."
        )
        if ref_root.text and test_root.text:
            assert (
                ref_root.text.strip() == test_root.text.strip()
            ), "Got differen element text."
        if len(ref_root) > 0 and len(test_root) > 0:
            for x, y in zip(ref_root, test_root):
                recursive_compare(x, y)

    recursive_compare(ref_root, test_root)

    os.remove(test_xml)


def test_conversion():
    """
    Test conversion of NXentry
    """
    root = Path(__file__).parent / "data" / "NXentry.nxdl.xml"
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [str(root)])
    assert result.exit_code == 0
    # Replace suffixes
    yaml = root.parent / Path(root.with_suffix("").stem + "_parsed.yaml")
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [str(yaml)])
    assert result.exit_code == 0
    new_root = yaml.with_suffix(".nxdl.xml")
    with open(root, encoding="utf-8", mode="r") as tmp_f:
        root_content = tmp_f.readlines()
    with open(new_root, encoding="utf-8", mode="r") as tmp_f:
        new_root_content = tmp_f.readlines()
    assert root_content == new_root_content
    Path.unlink(yaml)
    Path.unlink(new_root)


def test_yaml2nxdl_doc():
    """To test the doc style from yaml to nxdl."""
    pwd = Path(__file__).parent

    doc_file = pwd / "data/doc_yaml2nxdl.yaml"
    ref_doc_file = pwd / "data/ref_doc_yaml2nxdl.nxdl.xml"
    out_doc_file = (
        pwd / "data/doc_yaml2nxdl.nxdl.xml"
    )  # doc_file.with_suffix('.nxdl.xml')
    # Test yaml2nxdl
    # Generates '../data/doc_text.nxdl.xml'
    result = CliRunner().invoke(nyaml2nxdl.launch_tool, [str(doc_file)])
    if result.exit_code != 0:
        Path.unlink(out_doc_file)
    assert result.exit_code == 0, f"Error: Having issue running input file {doc_file}."

    ref_nxdl = ET.parse(str(ref_doc_file)).getroot()
    out_nxdl = ET.parse(str(out_doc_file)).getroot()

    def compare_nxdl_doc(parent1, parent2):
        if len(parent1) > 0 and len(parent2) > 0:
            for par1, par2 in zip(parent1, parent2):
                compare_nxdl_doc(par1, par2)

        elif (
            remove_namespace_from_tag(parent1.tag) == "doc"
            and remove_namespace_from_tag(parent2.tag) == "doc"
        ):
            assert (
                parent1.text == parent2.text
            ), f"DOCS ARE NOT SAME: node {parent1}, node {parent2}"

    compare_nxdl_doc(ref_nxdl, out_nxdl)

    Path.unlink(out_doc_file)


def test_nxdl2yaml_doc():
    """To test the doc style from nxdl to yaml."""

    pwd = Path(__file__).parent
    nxdl_file = pwd / "data/doc_nxdl2yaml.nxdl.xml"
    ref_yaml = pwd / "data/ref_doc_nxdl2yaml.yaml"
    parsed_yaml_file = pwd / "data/doc_nxdl2yaml_parsed.yaml"

    result = CliRunner().invoke(
        nyaml2nxdl.launch_tool, ["--do-not-store-nxdl", str(nxdl_file)]
    )

    if result.exit_code != 0:
        Path.unlink(parsed_yaml_file)

    assert result.exit_code == 0, "Error in converter execution."

    with open(ref_yaml, mode="r", encoding="utf-8") as yaml1, open(
        parsed_yaml_file, mode="r", encoding="utf-8"
    ) as yaml2:
        yaml_dict1 = LineLoader(yaml1).get_single_data()
        yaml_dict2 = LineLoader(yaml2).get_single_data()

    def compare_yaml_doc(yaml_dict1, yaml_dict2):
        for k_val1, k_val2 in zip(yaml_dict1.items(), yaml_dict2.items()):
            key1, val1 = k_val1
            key2, val2 = k_val2
            if key1 == "doc" and key2 == "doc":
                assert val1 == val2, "Doc texts are not the same."
            elif isinstance(val1, dict) and isinstance(val2, dict):
                compare_yaml_doc(val1, val2)

    compare_yaml_doc(yaml_dict1, yaml_dict2)
    Path.unlink(parsed_yaml_file)


def test_nyaml_dim_keyword(tmp_path):
    """
    The dim keyword is correctly translated into nxdl with rank and dim.
    """

    pwd = Path(__file__).parent
    input_file = pwd / "data/dim_keyword.yaml"
    ref_file = pwd / "data/ref_dim_keyword.nxdl.xml"
    parsed_file = tmp_path / "dim_keyword.nxdl.xml"

    result = CliRunner().invoke(
        nyaml2nxdl.launch_tool,
        ["--do-not-store-nxdl", str(input_file), "--output-file", str(parsed_file)],
    )

    assert result.exit_code == 0, "Error in converter execution."
    assert ref_file.read_text() == parsed_file.read_text()


@pytest.mark.parametrize(
    "test_input,output,is_valid",
    [
        (
            """
    xref:
        spec: <spec>
        term: <term>
        url: <url>
    """,
            "    This concept is related to term `<term>`_ "
            "of the <spec> standard.\n\n.. _<term>: <url>",
            True,
        ),
        (
            """
    xref:
        spec: <spec>
         term: <term>
        url: <url>
    """,
            "Found invalid xref. Please make sure that your xref entries are valid yaml.",
            False,
        ),
        (
            """
    xref:
        spec: <spec>
        term: <term>
        url: <url>
        term: <term2>
    """,
            "Invalid xref. It contains nested or duplicate keys.",
            False,
        ),
        (
            """
    xref:
        spec: <spec>
        term: <term>
        url: <url>
        hallo: <term2>
    """,
            "Invalid xref. Too many keys.",
            False,
        ),
        (
            """
    xref:
        spec: <spec>
        my_key: <term>
        url: <url>
    """,
            "Invalid xref key `my_key`. Must be one of `term`, `spec` or `url`.",
            False,
        ),
        (
            """
    xref:
        spec: <spec>
        term:
            test: <nested_value>
        url: <url>
    """,
            "Invalid xref. It contains nested or duplicate keys.",
            False,
        ),
    ],
)
def test_handle_xref(test_input, output, is_valid):
    """
    Tests whether the xref generates a correct docstring.
    """
    if is_valid:
        assert handle_each_part_doc(test_input) == output
        return

    with pytest.raises(ValueError) as err:
        handle_each_part_doc(test_input)

    assert output == err.value.args[0]



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Collect comments in a list by CommentCollector class and each comment is an instance of Comment
class. Each `Comment` instance(sometimes refered as 'comment block') consists of text and line
info or neighbours info where the comment must be placed.

The class Comment is an abstract class for general functions or method to be implemented
XMLComment and YAMLComment class.
"""


from typing import Any, Dict, List, Tuple, Type, Union

from nyaml.helper import LineLoader

__all__ = ["Comment", "CommentCollector", "XMLComment", "YAMLComment"]


# pylint: disable=inconsistent-return-statements
class CommentCollector:
    """CommentCollector will store a full comment ('Comment') object in
    _comment_chain.
    """

    def __init__(self, input_file: str = None, loaded_obj: Union[object, Dict] = None):
        """
        Initialise CommentCollector
        parameters:
            input_file: raw input file (xml, yml)
            loaded_obj: file loaded by third party library
        """
        self._comment_chain: List = []
        self.file = input_file
        self._comment_tracker = 0
        self._comment_hash: Dict[Tuple, Type[Comment]] = {}
        self.comment: Type[Comment]
        if self.file and not loaded_obj:
            if self.file.endswith(".xml"):
                self.comment = XMLComment
            elif self.file.split(".")[-1] == "yaml":
                self.comment = YAMLComment
                with open(self.file, "r", encoding="utf-8") as plain_text_yaml:
                    loader = LineLoader(plain_text_yaml)
                    self.comment.__yaml_dict__ = loader.get_single_data()
            else:
                raise ValueError("Input file must be a 'yaml' or 'nxdl.xml' type.")
        elif self.file and loaded_obj:
            if self.file.split(".")[-1] == "yaml" and isinstance(loaded_obj, dict):
                self.comment = YAMLComment
                self.comment.__yaml_dict__ = loaded_obj
            else:
                raise ValueError(
                    "Incorrect inputs for CommentCollector e.g. Wrong file extension."
                )

        else:
            raise ValueError("Incorrect inputs for CommentCollector")

    def extract_all_comment_blocks(self):
        """Collect all comments.

        Note here that comment means (comment text + element or line info
        intended for comment.
        """
        id_ = 0
        single_comment = self.comment(comment_id=id_)
        with open(self.file, mode="r", encoding="UTF-8") as enc_f:
            lines = enc_f.readlines()
            # Make an empty line for last comment if no empty lines in original file
            if lines[-1] != "":
                lines.append("")
            end_line_num = len(lines) - 1
            for line_num, line in enumerate(lines):
                if single_comment.is_storing_single_comment():
                    # If the last comment comes without post nxdl fields, groups and attributes
                    if "++ SHA HASH ++" in line:
                        # Handle with stored nxdl.xml file that is not part of yaml
                        single_comment.process_each_line("post_comment", (line_num + 1))
                        self._comment_chain.append(single_comment)
                        break
                    if line_num < end_line_num:
                        # Processing file from Line number 1
                        single_comment.process_each_line(line, (line_num + 1))
                    else:
                        # For processing last line of file
                        single_comment.process_each_line(
                            line + "post_comment", (line_num + 1)
                        )
                        self._comment_chain.append(single_comment)
                else:
                    self._comment_chain.append(single_comment)
                    single_comment = self.comment(last_comment=single_comment)
                    single_comment.process_each_line(line, (line_num + 1))

    def get_next_comment(self):
        """
        Return comment from comment_chain that must come earlier in order.
        """
        return self._comment_chain[self._comment_tracker]

    def get_comment_by_line_info(self, comment_locs: Tuple[str, Union[int, str]]):
        """
        Get comment using line information.
        """
        if comment_locs in self._comment_hash:
            return self._comment_hash[comment_locs]

        line_annot, line_loc = comment_locs
        for cmnt in self._comment_chain:
            if line_annot in cmnt:
                line_loc_ = cmnt.get_line_number(line_annot)
                if line_loc == line_loc_:
                    self._comment_hash[comment_locs] = cmnt
                    return cmnt

    def remove_comment(self, ind):
        """Remove a comment from comment list."""
        if ind < len(self._comment_chain):
            del self._comment_chain[ind]
        else:
            raise ValueError("Oops! Index is out of range.")

    def reload_comment(self):
        """
        Update self._comment_tracker after done with last comment.
        """
        self._comment_tracker += 1

    def __contains__(self, comment_locs: tuple):
        """
        Confirm wether the comment corresponds to key_line and line_loc
            is exist or not.
            comment_locs is equvalant to (line_annotation, line_loc) e.g.
            (__line__doc and 35)
        """
        if not isinstance(comment_locs, tuple):
            raise TypeError(
                "Comment_locs should be 'tuple' containing line annotation "
                "(e.g.__line__doc) and line_loc (e.g. 35)."
            )
        line_annot, line_loc = comment_locs
        for cmnt in self._comment_chain:
            if line_annot in cmnt:
                line_loc_ = cmnt.get_line_number(line_annot)
                if line_loc == line_loc_:
                    self._comment_hash[comment_locs] = cmnt
                    return True
        return False

    def __getitem__(self, ind):
        """Get comment from  self.obj._comment_chain by index."""
        if isinstance(ind, int):
            if ind >= len(self._comment_chain):
                raise IndexError(
                    f"Oops! Comment index {ind} in {__class__} is out of range!"
                )
            return self._comment_chain[ind]

        if isinstance(ind, slice):
            start_n = ind.start or 0
            end_n = ind.stop or len(self._comment_chain)
            return self._comment_chain[start_n:end_n]

    def __iter__(self):
        """get comment ieratively"""
        return iter(self._comment_chain)


# pylint: disable=too-many-instance-attributes
class Comment:
    """
    This class is building yaml comment and the intended line for what comment is written.
    """

    def __init__(self, comment_id: int = -1, last_comment: "Comment" = None) -> None:
        """Comment object can be considered as a block element that includes
        document element (an entity for what the comment is written).
        """
        self._elemt: Dict[str, Any] = {}
        self._elemt_text: str = None
        self._is_elemt_found: bool = None
        self._is_elemt_stored: bool = None

        self._comnt: str = ""
        # If Multiple comments for one element or entity
        self._comnt_list: List[str] = []
        self.last_comment: "Comment" = last_comment if last_comment else None
        if comment_id >= 0 and last_comment:
            self.cid = comment_id
            self.last_comment = last_comment
        elif comment_id == 0 and not last_comment:
            self.cid = comment_id
            self.last_comment = None
        elif last_comment:
            self.cid = self.last_comment.cid + 1
            self.last_comment = last_comment
        else:
            raise ValueError("Neither last comment nor comment id dound")
        self._comnt_start_found: bool = False
        self._comnt_end_found: bool = False
        self.is_storing_single_comment = lambda: not (
            self._comnt_end_found and self._is_elemt_stored
        )

    def get_comment_text_list(self) -> Union[List, str]:
        """
        Extract comment text from entrire comment (comment text + elment or
        line for what comment is intended)
        """

    def append_comment(self, text: str) -> None:
        """
        Append lines of the same comment.
        """

    def store_element(self, args) -> None:
        """
        Strore comment text and line or element that is intended for comment.
        """


class XMLComment(Comment):
    """
    XMLComment to store xml comment element.
    """

    def __init__(self, comment_id: int = -1, last_comment: "Comment" = None) -> None:
        super().__init__(comment_id, last_comment)

    def process_each_line(self, text, line_num):
        """Take care of each line of text. Through which function the text
        must be passed should be decide here.
        """
        text = text.strip()
        if text and line_num:
            self.append_comment(text)
            if self._comnt_end_found and not self._is_elemt_found:
                # for multiple comment if exist
                if self._comnt:
                    self._comnt_list.append(self._comnt)
                    self._comnt = ""

            if self._comnt_end_found:
                self.store_element(text)

    def append_comment(self, text: str) -> None:
        # Comment in single line
        if "<!--" == text[0:4]:
            self._comnt_start_found = True
            self._comnt_end_found = False
            self._comnt = self._comnt + text.replace("<!--", "")
            if "-->" == text[-4:]:
                self._comnt_end_found = True
                self._comnt_start_found = False
                self._comnt = self._comnt.replace("-->", "")

        elif "-->" == text[0:4] and self._comnt_start_found:
            self._comnt_end_found = True
            self._comnt_start_found = False
            self._comnt = self._comnt + "\n" + text.replace("-->", "")
        elif self._comnt_start_found:
            self._comnt = self._comnt + "\n" + text

    # pylint: disable=arguments-differ, arguments-renamed
    def store_element(self, text) -> None:
        def collect_xml_attributes(text_part):
            for part in text_part:
                part = part.strip()
                if part and part.endswith('">'):
                    self._is_elemt_stored = True
                    self._is_elemt_found = False
                    part = part[0:-2]
                elif part and part.endswith('"/>'):
                    self._is_elemt_stored = True
                    self._is_elemt_found = False
                    part = part[0:-3]
                elif part and part.endswith("/>"):
                    self._is_elemt_stored = True
                    self._is_elemt_found = False
                    part = part[0:-2]
                elif part and part.endswith(">"):
                    self._is_elemt_stored = True
                    self._is_elemt_found = False
                    part = part[0:-1]
                elif part and part.endswith('"'):
                    part = part[0:-1]

                if '="' in part:
                    lf_prt, rt_prt = part.split('="')
                    if ":" in lf_prt:
                        continue
                else:
                    continue
                self._elemt[lf_prt] = str(rt_prt)

        if not self._elemt:
            self._elemt = {}
        # First check for comment part has been collected prefectly
        if text.startswith("</"):
            pass
        elif text.startswith("<") and not text.startswith("<!--"):
            self._is_elemt_found = True
            text = text.replace("<", "", 1)
            text_part = text.split(" ")
            # collect tag
            self._elemt["tag"] = text_part[0]
            self._elemt["attrib"] = {}
            collect_xml_attributes(text_part[1:])

        elif self._is_elemt_found:
            text_part = text.split(" ")
            collect_xml_attributes(text_part)

    def get_element_info(self):
        """
            The method returns info dict that includes:
        'tag' and 'attrib' keys.
        """
        return self._elemt

    def get_comment_text_list(self) -> Union[List, str]:
        """
        This method returns list of commnent text. As some xml element might have
        multiple separated comment intended for a single element.
        """
        return self._comnt_list


class YAMLComment(Comment):
    """
    This class for storing comment text as well as location of the comment e.g. line
    number of other in the file.
    NOTE:
     1. Do not delete any element from yaml dictionary (for loaded_obj. check: Comment_collector
     class. because this loaded file has been exploited in nyaml2nxdl forward tools.)
    """

    # Class level variable. The main reason behind that to follow structure of
    # abstract class 'Comment'
    __yaml_dict__: dict = {}
    __yaml_line_info: dict = {}
    __comment_escape_char = {"--": "-\\-"}

    def __init__(self, comment_id: int = -1, last_comment: "Comment" = None) -> None:
        """Initialization of YAMLComment follow Comment class."""
        super().__init__(comment_id, last_comment)
        self.collect_yaml_line_info(
            YAMLComment.__yaml_dict__, YAMLComment.__yaml_line_info
        )

    def process_each_line(self, text, line_num):
        """Process each line.

        Take care of each line of text. Through which function the text
        must be passed should be decide here.
        """
        text = text.strip()
        self.append_comment(text)
        if self._comnt_end_found and not self._is_elemt_found:
            if self._comnt:
                self._comnt_list.append(self._comnt)
                self._comnt = ""

        if self._comnt_end_found:
            line_key = ""
            ind = text.find(":")
            if ind > 0:
                line_key = "__line__" + text[0:ind]

            for l_num, l_key in self.__yaml_line_info.items():
                if line_num == int(l_num) and line_key == l_key:
                    self.store_element(line_key, line_num)
                    break
                # Comment comes very end of the file
                if text == "post_comment" and line_key == "":
                    line_key = "__line__post_comment"
                    self.store_element(line_key, line_num)

    def has_post_comment(self):
        """Ensure if this is a post comment or not.

        Post comment means the comment that comes at the very end without having any
        nxdl element(class, group, filed and attribute.).
        """
        for key, _ in self._elemt.items():
            if "__line__post_comment" == key:
                return True
        return False

    def append_comment(self, text: str) -> None:
        """Append comment texts to the associated comment.

            Collects all the line of the same comment and
        append them with that single comment.
        """
        # check for escape char
        text = self.replace_escape_char(text)
        # Empty line after last line of comment
        if not text and self._comnt_start_found:
            self._comnt_end_found = True
            self._comnt_start_found = False
        # For empty line inside doc or yaml file.
        elif not text:
            return
        elif text.startswith("# "):
            self._comnt_start_found = True
            self._comnt_end_found = False
            self._comnt = "" if not self._comnt else self._comnt + "\n"
            self._comnt = self._comnt + text[2:]
        elif text.startswith("#"):
            self._comnt_start_found = True
            self._comnt_end_found = False
            self._comnt = "" if not self._comnt else self._comnt + "\n"
            self._comnt = self._comnt + text[1:]
        elif "post_comment" == text:
            self._comnt_end_found = True
            self._comnt_start_found = False
        # for any line after 'comment block' found
        elif self._comnt_start_found:
            self._comnt_start_found = False
            self._comnt_end_found = True

    # pylint: disable=arguments-differ
    def store_element(self, line_key, line_number):
        """Store comment contents and information of comment location."""
        self._elemt = {}
        self._elemt[line_key] = int(line_number)
        self._is_elemt_found = False
        self._is_elemt_stored = True

    def get_comment_text_list(self):
        """
        Return list of comments.
        """
        return self._comnt_list

    def get_line_number(self, line_key):
        """
        Return line no for which line the comment is created.
        """
        return self._elemt[line_key]

    def get_line_info(self):
        """
        Return line annotation and line number from a comment.
        """
        for line_anno, line_loc in self._elemt.items():
            return line_anno, line_loc

    def replace_escape_char(self, text):
        """Replace escape char according to __comment_escape_char dict"""
        for ecp_char, ecp_alt in YAMLComment.__comment_escape_char.items():
            text = text.replace(ecp_char, ecp_alt)
        return text

    def get_element_location(self):
        """Return yaml line '__line__<KEYY>' info and and line numner"""
        if len(self._elemt) == 0:
            raise ValueError(
                f"Comment element should be one or more but got {self._elemt}"
            )
        return next(self._elemt.items())

    def collect_yaml_line_info(self, yaml_dict, line_info_dict):
        """Collect __line__key and corresponding value from
        a yaml file dictonary in another dictionary.
        """
        for line_key, line_n in yaml_dict.items():
            if "__line__" in line_key:
                line_info_dict[line_n] = line_key

        for _, val in yaml_dict.items():
            if isinstance(val, dict):
                self.collect_yaml_line_info(val, line_info_dict)

    def __contains__(self, line_key):
        """For checking whether __line__<NAME> is in _elemt dict or not."""
        return line_key in self._elemt

    def __eq__(self, comment_obj):
        """Check the self has same value as right comment."""
        if not isinstance(comment_obj, Comment):
            raise TypeError(f"Expecting comment_obj as a instance of {Comment}")
        if len(self._comnt_list) != len(comment_obj._comnt_list):
            return False
        for left_cmnt, right_cmnt in zip(self._comnt_list, comment_obj._comnt_list):
            left_cmnt = left_cmnt.split("\n")
            right_cmnt = right_cmnt.split("\n")
            for left_line, right_line in zip(left_cmnt, right_cmnt):
                if left_line.strip() != right_line.strip():
                    return False
        return True



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Main file of nyaml2nxdl tool.

To write a definition for a instrument, experiment and/or process in nxdl.xml file from a YAML
file which details a hierarchy of data/metadata elements. It also allows both wa
conversion beteen YAML and nxdl.xml files that follows rules of NeXus ontology or data format.
"""
from pathlib import Path

import click

from nyaml.helper import (
    extend_yamlfile_by_nxdl_as_comment,
    get_sha256_hash,
    separate_hash_yaml_and_nxdl,
)
from nyaml.nxdl2nyaml import Nxdl2yaml
from nyaml.nyaml2nxdl import nyaml2nxdl

DEPTH_SIZE = 4 * " "
NXDL_SUFFIX = ".nxdl.xml"

# NOTE: Some handful links for nyaml2nxdl converter:
# https://manual.nexusformat.org/nxdl_desc.html?highlight=optional


def generate_nxdl_or_retrieve_nxdl(yaml_file, out_xml_file, verbose):
    """
    Generate yaml, nxdl and hash.

    If the extracted hash is exactly the same as produced from generated yaml then
    retrieve the nxdl part from provided yaml.
    Else, generate nxdl from separated yaml with the help of nyaml2nxdl function
    """
    file_path = Path(yaml_file)
    pa_path, rel_file = file_path.parent, file_path.name
    sep_yaml = (pa_path / f"temp_{rel_file}").as_posix()
    hash_found = separate_hash_yaml_and_nxdl(yaml_file, sep_yaml, out_xml_file)

    if hash_found:
        gen_hash = get_sha256_hash(sep_yaml)
        if hash_found == gen_hash:
            Path(sep_yaml).unlink()
            return

    nyaml2nxdl(sep_yaml, out_xml_file, verbose)
    Path(sep_yaml).unlink()


def split_name_and_extension(file_path):
    """
    Split file name into extension and rest of the file name.

    return file raw name and extension
    """
    path = Path(file_path)
    ext = "".join(path.suffixes)
    full_path_stem = file_path[0 : file_path.index(ext)]
    return full_path_stem, ext[1:]


@click.command()
@click.argument("input-file")
@click.option(
    "--output-file",
    required=False,
    help="The output file path to write the converted file to",
)
@click.option(
    "--check-consistency",
    is_flag=True,
    default=False,
    help=(
        "Check if yaml and nxdl can be converted from one to another version recursively and"
        " get the same version of file. E.g. from NXexample.nxdl.xml"
        " to NXexample_consistency.nxdl.xml."
    ),
)
@click.option(
    "--do-not-store-nxdl",
    is_flag=True,
    default=False,
    help=(
        "Whether the input nxdl file will be stored as a comment"
        " at the end of output yaml file."
    ),
)
@click.option(
    "--verbose",
    is_flag=True,
    default=False,
    help="Print in standard output keywords and value types to help \
possible issues in yaml files",
)
# def launch_tool(input_file, verbose, check_consistency):
def launch_tool(input_file, verbose, do_not_store_nxdl, check_consistency, output_file):
    """
    Main function that distinguishes the input file format and launches the tools.
    """

    if Path(input_file).is_file():
        raw_name, ext = split_name_and_extension(input_file)
    else:
        raise ValueError("Need a valid input file.")
    if ext == "yaml":
        xml_out_file = (
            f"{raw_name}{NXDL_SUFFIX}" if output_file is None else output_file
        )
        generate_nxdl_or_retrieve_nxdl(input_file, xml_out_file, verbose)

        # For consistency running
        if check_consistency:
            yaml_out_file = f"{raw_name}_consistency.{ext}"
            converter = Nxdl2yaml([], [])
            converter.print_yml(xml_out_file, yaml_out_file, verbose)
            Path(xml_out_file).unlink()
    elif ext == "nxdl.xml":
        # if not append:
        yaml_out_file = (
            f"{raw_name}_parsed.yaml" if output_file is None else output_file
        )
        converter = Nxdl2yaml([], [])
        converter.print_yml(input_file, yaml_out_file, verbose)
        # Store nxdl.xml file in output yaml file under SHA HASH
        yaml_hash = get_sha256_hash(yaml_out_file)
        # Lines as divider between yaml and nxdl
        top_lines = [
            (
                "\n# ++++++++++++++++++++++++++++++++++ SHA HASH"
                " ++++++++++++++++++++++++++++++++++\n"
            ),
            f"# {yaml_hash}\n",
        ]
        if not do_not_store_nxdl:
            extend_yamlfile_by_nxdl_as_comment(
                yaml_file=yaml_out_file,
                file_to_be_appended=input_file,
                top_lines_list=top_lines,
            )

        # Taking care of consistency running
        if check_consistency:
            xml_out_file = f"{raw_name}_consistency.{ext}"
            generate_nxdl_or_retrieve_nxdl(yaml_out_file, xml_out_file, verbose)
            Path.unlink(yaml_out_file)
    else:
        raise ValueError("Provide correct file with extension '.yaml or '.nxdl.xml")


if __name__ == "__main__":
    launch_tool().parse()  # pylint: disable=no-value-for-parameter



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Creates an instantiated NXDL schema XML tree by walking the dictionary nest
"""

import datetime
import pathlib
import re
import textwrap
import warnings
from typing import Union
from urllib.parse import unquote

import lxml.etree as ET
import yaml
from yaml.scanner import ScannerError

from nyaml.comment_collector import CommentCollector
from nyaml.helper import (
    YAML_ATTRIBUTES_ATTRIBUTES,
    YAML_FIELD_ATTRIBUTES,
    YAML_GROUP_ATTRIBUTES,
    YAML_LINK_ATTRIBUTES,
    LineLoader,
    clean_empty_lines,
    get_yaml_escape_char_reverter_dict,
    is_dom_comment,
    nx_name_type_resolving,
    remove_namespace_from_tag,
)

# pylint: disable=too-many-lines, global-statement, invalid-name
DOM_COMMENT = (
    f"# NeXus - Neutron and X-ray Common Data Format\n"
    f"#\n"
    f"# Copyright (C) 2014-{datetime.date.today().year} "
    "NeXus International Advisory Committee (NIAC)\n"
    f"#\n"
    f"# This library is free software; you can redistribute it and/or\n"
    f"# modify it under the terms of the GNU Lesser General Public\n"
    f"# License as published by the Free Software Foundation; either\n"
    f"# version 3 of the License, or (at your option) any later version.\n"
    f"#\n"
    f"# This library is distributed in the hope that it will be useful,\n"
    f"# but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
    f"# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU\n"
    f"# Lesser General Public License for more details.\n"
    f"#\n"
    f"# You should have received a copy of the GNU Lesser General Public\n"
    f"# License along with this library; if not, write to the Free Software\n"
    f"# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"
    f"#\n"
    f"# For further information, see http://www.nexusformat.org\n"
)
NX_ATTR_IDNT = "\\@"
NX_UNIT_IDNT = "unit"
DEPTH_SIZE = 4 * " "
# Initialised in yml_reader() funtion
COMMENT_BLOCKS: CommentCollector
CATEGORY = ""  # Definition would be either 'base' or 'application'


def check_for_dom_comment_in_yaml():
    """Check the yaml file has dom comment or dom comment needed to be hard coded."""

    # Check for dom comments in first five comments
    dom_comment = ""
    dom_comment_ind = 1
    for ind, comnt in enumerate(COMMENT_BLOCKS[0:5]):
        cmnt_list = comnt.get_comment_text_list()
        if len(cmnt_list) == 1:
            text = cmnt_list[0]
        else:
            continue
        if is_dom_comment(text):
            dom_comment = text
            dom_comment_ind = ind
            break

    # deactivate the root dom_comment, So that the corresponding comment would not be
    # considered as comment for definition xml element.
    if dom_comment:
        COMMENT_BLOCKS.remove_comment(dom_comment_ind)

    return dom_comment


def yml_reader(inputfile):
    """
    This function launches the LineLoader class.
    It parses the yaml in a dict and extends it with line tag keys for each key of the dict.
    """
    global COMMENT_BLOCKS
    with open(inputfile, "r", encoding="utf-8") as plain_text_yaml:
        loader = LineLoader(plain_text_yaml)
        loaded_yaml = loader.get_single_data()
    COMMENT_BLOCKS = CommentCollector(inputfile, loaded_yaml)
    COMMENT_BLOCKS.extract_all_comment_blocks()
    dom_cmnt_frm_yaml = check_for_dom_comment_in_yaml()
    global DOM_COMMENT
    if dom_cmnt_frm_yaml:
        DOM_COMMENT = dom_cmnt_frm_yaml

    if "category" not in loaded_yaml.keys():
        raise ValueError(
            "All definitions should be either 'base' or 'application' category. "
            "No category has been found."
        )
    global CATEGORY
    CATEGORY = loaded_yaml["category"]
    return loaded_yaml


def check_for_default_attribute_and_value(xml_element):
    """Check for default attribute for NeXus concepts.

    NeXus Groups, fields and attributes might have xml default attributes and values that must
    come. For example: 'optional' which is 'true' by default for base class and false otherwise.
    """

    # base: Default attributes and value for all elements of base class except dimension element
    base_attr_to_val = {"optional": "true"}

    # application: Default attributes and value for all elements of application class except
    # dimension element
    application_attr_to_val = {"optional": "false"}

    # Default attributes and value for dimension element
    base_dim_attr_to_val = {"required": "false"}
    application_dim_attr_to_val = {"required": "true"}

    # Eligible tag for default attr and value
    eligible_tag = ["group", "field", "attribute"]

    def set_default_attribute(xml_elem, default_attr_to_val):
        for deflt_attr, deflt_val in default_attr_to_val.items():
            if (
                deflt_attr not in xml_elem.attrib
                and "maxOccurs" not in xml_elem.attrib
                and "minOccurs" not in xml_elem.attrib
                and "recommended" not in xml_elem.attrib
            ):
                xml_elem.set(deflt_attr, deflt_val)

    for child in list(xml_element):
        # skipping comment 'function' that mainly collect comment from yaml file.
        if not isinstance(child.tag, str):
            continue
        tag = remove_namespace_from_tag(child.tag)

        if tag == "dim" and CATEGORY == "base":
            set_default_attribute(child, base_dim_attr_to_val)
        if tag == "dim" and CATEGORY == "application":
            set_default_attribute(child, application_dim_attr_to_val)
        if tag in eligible_tag and CATEGORY == "base":
            set_default_attribute(child, base_attr_to_val)
        if tag in eligible_tag and CATEGORY == "application":
            set_default_attribute(child, application_attr_to_val)
        check_for_default_attribute_and_value(child)


def yml_reader_nolinetag(inputfile):
    """
    pyyaml based parsing of yaml file in python dict
    """
    with open(inputfile, "r", encoding="utf-8") as stream:
        parsed_yaml = yaml.safe_load(stream)
    return parsed_yaml


def check_for_skipped_attributes(component, value, allowed_attr=None, verbose=False):
    """
    Check for any attributes have been skipped or not.
    NOTE: We should keep in mind about 'doc'
    """
    block_tag = ["enumeration"]
    if value:
        for attr, val in value.items():
            if attr == "doc":
                continue
            if "__line__" in attr or attr in block_tag:
                continue
            line_number = f"__line__{attr}"
            if verbose:
                print(f"__line__ : {value[line_number]}")
            if (
                not isinstance(val, dict)
                and "\\@" not in attr
                and attr not in allowed_attr
                and "NX" not in attr
                and val
            ):
                raise ValueError(
                    f"An attribute '{attr}' in part '{component}' has been found"
                    f". Please check around line '{value[line_number]}. At this "
                    f"time, the allowed attrbutes are {allowed_attr}."
                )


def format_nxdl_doc(string):
    """NeXus format for doc string"""
    string = check_for_mapping_char_other(string)
    string_len = 80
    if "\n" not in string:
        if len(string) > string_len:
            wrapped = textwrap.TextWrapper(
                width=string_len, break_long_words=False, replace_whitespace=False
            )
            string = "\n".join(wrapped.wrap(string))
        formatted_doc = "\n" + f"{string}"
    else:
        text_lines = string.split("\n")
        text_lines = clean_empty_lines(text_lines)
        formatted_doc = "\n" + "\n".join(text_lines)
    if not formatted_doc.endswith("\n"):
        formatted_doc += "\n"
    return formatted_doc


def check_for_mapping_char_other(text):
    """
    Check for mapping char \':\' which does not be passed through yaml library.
    Then replace it by ':'.
    """
    if not text:
        text = ""
    text = str(text)
    if text == "True":
        text = "true"
    if text == "False":
        text = "false"
    # Some escape char is not valid in yaml libray which is written while writing
    # yaml file. In the time of writing nxdl revert to that escape char.
    escape_reverter = get_yaml_escape_char_reverter_dict()
    for key, val in escape_reverter.items():
        if key in text:
            text = text.replace(key, val)
    return text.strip()


def handle_each_part_doc(text):
    """Check and handle if the text is corresponds to xref or plain doc.

    In nyaml doc the entire documentation may come in list of small docs.
    one doc string might be as follows:
    '''
    xref:
        spec: <spec>
        term: <term>
        url: <url>
    '''

    which has to be formatted as
    '''
        This concept is related to term `<term>`_ of the <spec> standard.
    .. _<term>: <url>


    Parameters
    ----------
    text : string
        String that looks like yaml notaion.

    return
    ------
    Formated text
    """

    clean_txt = text.strip()

    if not clean_txt.startswith("xref:"):
        return format_nxdl_doc(check_for_mapping_char_other(clean_txt)).strip()

    no_lines = len(clean_txt.splitlines())
    try:
        xref_dict = yaml.safe_load(clean_txt)
    except ScannerError as scan_err:
        raise ValueError(
            "Found invalid xref. Please make sure that your xref entries are valid yaml."
        ) from scan_err
    xref_entries = xref_dict.get("xref", {})

    if no_lines != len(xref_entries) + 1:
        raise ValueError("Invalid xref. It contains nested or duplicate keys.")

    if no_lines > 4:
        raise ValueError("Invalid xref. Too many keys.")

    for key in xref_entries:
        if key not in ("term", "spec", "url"):
            raise ValueError(
                f"Invalid xref key `{key}`. Must be one of `term`, `spec` or `url`."
            )

    return (
        f"    This concept is related to term `{xref_entries.get('term', 'NO TERM')}`_ "
        f"of the {xref_entries.get('spec', 'NO TERM')} standard.\n\n"
        f".. _{xref_entries.get('term', 'NO SPECIFICATION')}: "
        f"{xref_entries.get('url', 'NO URL')}"
    )


def xml_handle_doc(obj, value: Union[str, list], line_number=None, line_loc=None):
    """This function creates a 'doc' element instance, and appends it to an existing element"""
    # global comment_bolcks
    doc_elemt = ET.SubElement(obj, "doc")
    text = ""
    if isinstance(value, list):
        for doc_part in value:
            text = text + "\n" + handle_each_part_doc(doc_part) + "\n"
    else:
        text = text + "\n" + handle_each_part_doc(value) + "\n"
    # To keep the doc middle of doc tag.
    doc_elemt.text = text
    if line_loc is not None and line_number is not None:
        xml_handle_comment(obj, line_number, line_loc, doc_elemt)


def xml_handle_units(obj, value):
    """This function creates a 'units' element instance, and appends it to an existing element"""
    obj.set("units", str(value))


# pylint: disable=too-many-branches
def xml_handle_exists(dct, obj, keyword, value):
    """
    This function creates an 'exists' element instance, and appends it to an existing element
    """
    line_number = f"__line__{keyword}"
    assert (
        value is not None
    ), f"Line {dct[line_number]}: exists argument must not be None !"
    if isinstance(value, list):
        if len(value) == 4:
            if value[0] == "min" and value[2] == "max":
                obj.set("minOccurs", str(value[1]))
                if str(value[3]) != "infty":
                    obj.set("maxOccurs", str(value[3]))
                else:
                    obj.set("maxOccurs", "unbounded")
            elif value[0] == "max" and value[2] == "min":
                if str(value[1]) != "infty":
                    obj.set("maxOccurs", str(value[1]))
                else:
                    obj.set("maxOccurs", "unbounded")
                obj.set("minOccurs", str(value[3]))
            else:
                raise ValueError(
                    f"Line {dct[line_number]}: exists keyword"
                    f"needs to go either with an optional [recommended] list with two "
                    f"entries either [min, <uint>] or [max, <uint>], or a list of four "
                    f"entries [min, <uint>, max, <uint>] !"
                )
        elif len(value) == 2 and value[0] == "min":
            obj.set("minOccurs", str(value[1]))
        elif len(value) == 2 and value[0] == "max":
            obj.set("maxOccurs", str(value[1]))
        else:
            raise ValueError(
                f"Line {dct[line_number]}: exists keyword "
                f"needs to go either with optional, recommended, a list with two "
                f"entries either [min, <uint>] or [max, <uint>], or a list of four "
                f"entries [min, <uint>, max, <uint>] !"
            )
    else:
        # This clause take optional in all cases except dimension where 'required' key is allowed
        # not the 'optional' key.
        if value == "optional":
            obj.set("optional", "true")
        elif value == "recommended":
            obj.set("recommended", "true")
        elif value == "required":
            obj.set("optional", "false")
        else:
            obj.set("minOccurs", "0")


def xml_handle_dimensions(dct, obj, keyword, value: dict):
    """
    This function creates a 'dimensions' element instance, and appends it to an existing element

    NOTE: we could create xml_handle_dim() function.
        But, the dim elements in yaml file is defined as 'dim =[[index, value]]'
        but dim has other attributes such as 'ref' and also might have doc as chlid.
        so in that sense 'dim' should have come as dict keeping attributes and child as members of
        dict.
        Regarding this situation all the attributes of 'dimensions' and child 'doc' has been
        included here.

        Other attributes, except 'index' and 'value', of 'dim' comes under nested dict named
        'dim_parameter:
            incr:[...]'
    """

    possible_dimension_attrs = ["rank"]  # nxdl attributes
    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    assert (
        "dim" in value.keys()
    ), f"Line {line_loc}: No dim as child of dimension has been found."
    xml_handle_comment(obj, line_number, line_loc)
    dims = ET.SubElement(obj, "dimensions")
    # Consider all the children under dimension is dim element and
    # its attributes

    rm_key_list = []
    rank = ""
    for key, val in value.items():
        if "__line__" in key:
            continue
        line_number = f"__line__{key}"
        line_loc = value[line_number]
        if key == "rank":
            rank = val or ""
            if isinstance(rank, int) and rank < 0:
                raise ValueError(
                    f"Dimension must have some info about rank which is not "
                    f"available. Please check around Line: {dct[line_number]}"
                )
            dims.set(key, str(val))
            rm_key_list.append(key)
            rm_key_list.append(line_number)
            xml_handle_comment(obj, line_number, line_loc, dims)
        # Check dimension doc and handle it
        elif key == "doc" and isinstance(val, str):
            xml_handle_doc(dims, val, line_number, line_loc)
            rm_key_list.append(key)
            rm_key_list.append(line_number)
        elif key in possible_dimension_attrs and not isinstance(val, dict):
            dims.set(key, str(val))
            rm_key_list.append(key)
            rm_key_list.append(line_number)
            xml_handle_comment(obj, line_number, line_loc, dims)

    for key in rm_key_list:
        del value[key]

    xml_handle_dim_from_dimension_dict(dct, dims, keyword, value, rank=False)

    if isinstance(value, dict) and value != {}:
        recursive_build(dims, value, verbose=None)


def xml_handle_dim(obj, value):
    """
    This function creates a 'dimensions' element instance, and appends it to an existing element.
    Allows for handling numpy tensor notation of dimensions. That is,
    dimensions:
      rank: 1
      dim: (1, 3)
    can be replaced by
    dim: (3,)
    """
    if isinstance(value, str):
        if value[0] == "(" and value[-1] == ")":
            valid_dims = []
            for entry in value[1:-1].replace(" ", "").split(","):
                if len(entry) > 0:  # ignore trailing comma and empty mnemonics
                    valid_dims.append(entry)
            if len(valid_dims) > 0:
                dims = ET.SubElement(obj, "dimensions")
                dims.set("rank", str(len(valid_dims)))
                dim_idx = 1
                for dim_name in valid_dims:
                    dim = ET.SubElement(dims, "dim")
                    dim.set("index", str(dim_idx))
                    dim.set("value", str(dim_name))
                    dim_idx += 1


# pylint: disable=too-many-locals, too-many-arguments, too-many-statements
def xml_handle_dim_from_dimension_dict(
    dct, dims_obj, keyword, value, rank, verbose=False
):
    """
    Handling dim element.
    NOTE: The inputs 'keyword' and 'value' are as input for xml_handle_dimensions
    function. please also read note in xml_handle_dimensions.
    """
    deprecated_dim_attrs = ["ref", "incr", "refindex"]
    possible_dim_attrs = [*deprecated_dim_attrs, "required"]

    # Some attributes might have equivalent name e.g. 'required' is correct one and
    # 'optional' could be another name. Then change attribute to the correct one.
    wrong_to_correct_attr = [("optional", "required")]
    header_line_number = f"__line__{keyword}"
    dim_list = []
    rm_key_list = []
    # NOTE: dim doc and other attributes except 'index' and 'value' will come as list of value
    # under dim_parameters
    if not value:
        return
    rank = ""
    # pylint: disable=too-many-nested-blocks
    for attr, vvalue in value.items():
        if "__line__" in attr:
            continue
        line_number = f"__line__{attr}"
        line_loc = value[line_number]
        # dim comes in precedence
        if attr == "dim":
            # dim consists of [index, value] list
            llist_ind_value = vvalue
            assert isinstance(
                llist_ind_value, list
            ), f"Line {value[line_number]}: dim argument not a list !"
            xml_handle_comment(dims_obj, line_number, line_loc)
            if isinstance(rank, int) and rank > 0:
                assert rank == len(llist_ind_value), (
                    f"Wrong dimension rank check around Line {dct[header_line_number]}.\n"
                    f"Line {[dct[header_line_number]]} rank value {rank} "
                    f"is not the same as dim array = "
                    f"{len(llist_ind_value)}."
                )
            # Taking care of ind and value that comes as list of list
            for dim_ind_val in llist_ind_value:
                dim = ET.SubElement(dims_obj, "dim")

                # Taking care of multidimensions or rank
                if len(dim_ind_val) >= 1 and dim_ind_val[0]:
                    dim.set("index", str(dim_ind_val[0]))
                if len(dim_ind_val) == 2 and dim_ind_val[1]:
                    dim.set("value", str(dim_ind_val[1]))
                dim_list.append(dim)
            rm_key_list.append(attr)
            rm_key_list.append(line_number)
        elif attr == "dim_parameters" and isinstance(vvalue, dict):
            xml_handle_comment(dims_obj, line_number, line_loc)
            for kkkey, vvval in vvalue.items():
                if "__line__" in kkkey:
                    continue
                cmnt_number = f"__line__{kkkey}"
                cmnt_loc = vvalue[cmnt_number]
                # Check whether any optional attributes added
                for tuple_wng_crt in wrong_to_correct_attr:
                    if kkkey == tuple_wng_crt[0]:
                        raise ValueError(
                            f"{cmnt_loc}: Attribute '{kkkey}' is prohibited, use "
                            f"'{tuple_wng_crt[1]}"
                        )
                if kkkey == "doc" and dim_list:
                    # doc comes as list of doc
                    for i, dim in enumerate(dim_list):
                        if isinstance(vvval, list) and i < len(vvval):
                            tmp_val = vvval[i]
                            xml_handle_doc(dim, vvval[i], cmnt_number, cmnt_loc)
                        # Check all the dim have doc if not skip
                        elif isinstance(vvval, list) and i >= len(vvval):
                            pass
                else:
                    if kkkey in deprecated_dim_attrs:
                        dep_text = (
                            f"Attrbute {kkkey} is deprecated. "
                            f"Check attributes after line {cmnt_loc}"
                        )
                        warnings.warn(dep_text, DeprecationWarning)
                    for i, dim in enumerate(dim_list):
                        # all atribute of dims comes as list
                        if isinstance(vvval, list) and i < len(vvval):
                            tmp_val = vvval[i]
                            dim.set(kkkey, str(tmp_val))

                        # Check all the dim have doc if not skip
                        elif isinstance(vvval, list) and i >= len(vvval):
                            pass
                        # All dim might have the same value for the same attribute
                        elif not isinstance(vvval, list):
                            tmp_val = value
                            dim.set(kkkey, str(tmp_val))
            rm_key_list.append(attr)
            rm_key_list.append(line_number)
        else:
            raise ValueError(
                f"Got unexpected block except 'dim' and 'dim_parameters'."
                f"Please check around line {line_number}"
            )

    for key in rm_key_list:
        del value[key]

    check_for_skipped_attributes("dim", value, possible_dim_attrs, verbose)


def xml_handle_enumeration(dct, obj, keyword, value, verbose):
    """This function creates an 'enumeration' element instance.

    Two cases are handled:
    1) the items are in a list
    2) the items are dictionaries and may contain a nested doc
    """
    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    xml_handle_comment(obj, line_number, line_loc)
    enum = ET.SubElement(obj, "enumeration")

    assert (
        value is not None
    ), f"Line {line_loc}: enumeration must \
bear at least an argument !"
    assert (
        len(value) >= 1
    ), f"Line {dct[line_number]}: enumeration must not be an empty list!"
    if isinstance(value, list):
        for element in value:
            itm = ET.SubElement(enum, "item")
            itm.set("value", str(element))
    if isinstance(value, dict) and value != {}:
        for element in value.keys():
            if "__line__" not in element:
                itm = ET.SubElement(enum, "item")
                itm.set("value", str(element))
                if isinstance(value[element], dict):
                    recursive_build(itm, value[element], verbose)


# pylint: disable=unused-argument
def xml_handle_link(dct, obj, keyword, value, verbose):
    """
    If we have an NXDL link we decode the name attribute from <optional string>(link)[:-6]
    """

    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    xml_handle_comment(obj, line_number, line_loc)
    name = keyword[:-6]
    link_obj = ET.SubElement(obj, "link")
    link_obj.set("name", str(name))

    if value:
        rm_key_list = []
        for attr, vval in value.items():
            if "__line__" in attr:
                continue
            line_number = f"__line__{attr}"
            line_loc = value[line_number]
            if attr == "doc":
                xml_handle_doc(link_obj, vval, line_number, line_loc)
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
            elif attr in YAML_LINK_ATTRIBUTES and not isinstance(vval, dict):
                if vval:
                    link_obj.set(attr, str(vval))
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
                xml_handle_comment(obj, line_number, line_loc, link_obj)

        for key in rm_key_list:
            del value[key]
        # Check for skipped attributes
        check_for_skipped_attributes("link", value, YAML_LINK_ATTRIBUTES, verbose)

    if isinstance(value, dict) and value != {}:
        recursive_build(link_obj, value, verbose=None)


def xml_handle_choice(dct, obj, keyword, value, verbose=False):
    """
    Build choice xml elements. That consists of groups.
    """
    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    xml_handle_comment(obj, line_number, line_loc)
    # Add to this tuple if new attributes have been added to nexus definition.
    possible_attr = ()
    choice_obj = ET.SubElement(obj, "choice")
    # take care of special attributes
    name = keyword[:-8]
    choice_obj.set("name", name)

    if value:
        rm_key_list = []
        for attr, vval in value.items():
            if "__line__" in attr:
                continue
            line_number = f"__line__{attr}"
            line_loc = value[line_number]
            if attr == "doc":
                xml_handle_doc(choice_obj, vval, line_number, line_loc)
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
            elif attr in possible_attr and not isinstance(vval, dict):
                if vval:
                    choice_obj.set(attr, str(vval))
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
                xml_handle_comment(obj, line_number, line_loc, choice_obj)

        for key in rm_key_list:
            del value[key]
        # Check for skipped attributes
        check_for_skipped_attributes("choice", value, possible_attr, verbose)

    if isinstance(value, dict) and value != {}:
        recursive_build(choice_obj, value, verbose=None)


def xml_handle_symbols(dct, obj, keyword, value: dict):
    """Handle a set of NXDL symbols as a child to obj"""
    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    assert (
        len(list(value.keys())) > 0
    ), f"Line {line_loc}: symbols table must not be empty !"
    xml_handle_comment(obj, line_number, line_loc)
    syms = ET.SubElement(obj, "symbols")
    if "doc" in value.keys():
        line_number = "__line__doc"
        line_loc = value[line_number]
        xml_handle_comment(syms, line_number, line_loc)
        doctag = ET.SubElement(syms, "doc")
        doctag.text = "\n" + textwrap.fill(value["doc"], width=70) + "\n"
    rm_key_list = []
    for kkeyword, vvalue in value.items():
        if "__line__" in kkeyword:
            continue
        if kkeyword != "doc":
            line_number = f"__line__{kkeyword}"
            line_loc = value[line_number]
            xml_handle_comment(syms, line_number, line_loc)
            assert vvalue is not None and isinstance(
                vvalue, str
            ), f"Line {line_loc}: put a comment in doc string !"
            sym = ET.SubElement(syms, "symbol")
            sym.set("name", kkeyword)
            xml_handle_doc(sym, vvalue)
            rm_key_list.append(kkeyword)
            rm_key_list.append(line_number)
    for key in rm_key_list:
        del value[key]


def check_keyword_variable(verbose, dct, keyword, value):
    """
    Check whether both keyword_name and keyword_type are empty,
        and complains if it is the case
    """
    keyword_name, keyword_type = nx_name_type_resolving(keyword)
    if verbose:
        print(f"{keyword_name}({keyword_type}): value type is {type(value)}\n")
    if keyword_name == "" and keyword_type == "":
        line_number = f"__line__{keyword}"
        raise ValueError(f"Line {dct[line_number]}: found an improper yaml key !")


def helper_keyword_type(kkeyword_type):
    """
    Return a value of keyword_type if it belong to NX_TYPE_KEYS
    """
    if re.match(r"NX_[A-Z]+", kkeyword_type):
        return kkeyword_type
    return None


def verbose_flag(verbose, keyword, value):
    """
    Verbose stdout printing for nested levels of yaml file, if verbose flag is active
    """
    if verbose:
        print(f"key:{keyword}; value type is {type(value)}\n")


def xml_handle_attributes(dct, obj, keyword, value, verbose):
    """Handle the attributes found connected to attribute field"""

    line_number = f"__line__{keyword}"
    line_loc = dct[line_number]
    xml_handle_comment(obj, line_number, line_loc)
    # as an attribute identifier
    keyword_name, keyword_typ = nx_name_type_resolving(keyword)
    line_number = f"__line__{keyword}"
    if verbose:
        print(f"__line__ : {dct[line_number]}")
    if keyword_name == "" and keyword_typ == "":
        raise ValueError(f"Line {dct[line_number]}: found an improper yaml key !")
    elemt_obj = ET.SubElement(obj, "attribute")
    elemt_obj.set("name", keyword_name[2:])
    if keyword_typ:
        elemt_obj.set("type", keyword_typ)

    rm_key_list = []
    if value and value:
        # taking care of attributes of attributes
        for attr, attr_val in value.items():
            if "__line__" in attr:
                continue
            line_number = f"__line__{attr}"
            line_loc = value[line_number]
            if attr in ["doc", *YAML_ATTRIBUTES_ATTRIBUTES] and not isinstance(
                attr_val, dict
            ):
                if attr == "unit":
                    elemt_obj.set(f"{attr}s", str(value[attr]))
                    rm_key_list.append(attr)
                    rm_key_list.append(line_number)
                    xml_handle_comment(obj, line_number, line_loc, elemt_obj)
                elif attr == "exists" and attr_val:
                    xml_handle_exists(value, elemt_obj, attr, attr_val)
                    rm_key_list.append(attr)
                    rm_key_list.append(line_number)
                    xml_handle_comment(obj, line_number, line_loc, elemt_obj)
                elif attr == "doc":
                    xml_handle_doc(
                        elemt_obj, format_nxdl_doc(attr_val), line_number, line_loc
                    )
                    rm_key_list.append(attr)
                    rm_key_list.append(line_number)
                else:
                    elemt_obj.set(attr, check_for_mapping_char_other(attr_val))
                    rm_key_list.append(attr)
                    rm_key_list.append(line_number)
                    xml_handle_comment(obj, line_number, line_loc, elemt_obj)

        for key in rm_key_list:
            del value[key]
        # Check cor skipped attribute
        check_for_skipped_attributes(
            "Attribute", value, YAML_ATTRIBUTES_ATTRIBUTES, verbose
        )
    if value:
        recursive_build(elemt_obj, value, verbose)


def validate_field_attribute_and_value(v_attr, vval, allowed_attribute, value):
    """
    Check for any attributes that comes with invalid name,
        and invalid value.
    """

    if not isinstance(vval, dict) and not str(vval):  # check for empty value
        line_number = f"__line__{v_attr}"
        raise ValueError(
            f"In a field a valid attrbute ('{v_attr}') found that is not stored."
            f" Please check around line {value[line_number]}"
        )

    # The below elements might come as child element
    skipped_child_name = ["doc", "dimension", "enumeration", "choice", "exists"]
    # check for invalid key or attributes
    if (
        v_attr not in [*skipped_child_name, *allowed_attribute]
        and "__line__" not in v_attr
        and not isinstance(vval, dict)
        and "(" not in v_attr  # skip only groups and field that has name and type
        and "\\@" not in v_attr
    ):  # skip nexus attributes
        line_number = f"__line__{v_attr}"
        raise ValueError(
            f"In a field or group a invalid attribute ('{v_attr}') or child has found."
            f" Please check around line {value[line_number]}."
        )


def xml_handle_fields_or_group(
    dct, obj, keyword, value, ele_type, allowed_attr, verbose=False
):
    """Handle a field or group in yaml file."""
    line_annot = f"__line__{keyword}"
    line_loc = dct[line_annot]
    xml_handle_comment(obj, line_annot, line_loc)
    l_bracket = -1
    r_bracket = -1
    if keyword.count("(") == 1:
        l_bracket = keyword.index("(")
    if keyword.count(")") == 1:
        r_bracket = keyword.index(")")

    keyword_name, keyword_type = nx_name_type_resolving(keyword)
    if ele_type == "field" and not keyword_name:
        raise ValueError(
            f"No name for NeXus {ele_type} has been found."
            f"Check around line:{line_loc}"
        )
    if not keyword_type and not keyword_name:
        raise ValueError(
            f"No name or type for NeXus {ele_type} has been found."
            f"Check around line: {line_loc}"
        )

    elemt_obj = ET.SubElement(obj, ele_type)

    # type come first
    if l_bracket == 0 and r_bracket > 0:
        elemt_obj.set("type", keyword_type)
        if keyword_name:
            elemt_obj.set("name", keyword_name)
    elif l_bracket > 0:
        elemt_obj.set("name", keyword_name)
        if keyword_type:
            elemt_obj.set("type", keyword_type)
    else:
        elemt_obj.set("name", keyword_name)

    if value:
        rm_key_list = []
        # In each each if clause apply xml_handle_comment(), to collect
        # comments on that yaml line.
        for attr, vval in value.items():
            if "__line__" in attr:
                continue
            line_number = f"__line__{attr}"
            line_loc = value[line_number]
            if attr == "doc":
                xml_handle_doc(
                    elemt_obj,
                    vval,
                    line_number,
                    line_loc,
                )
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
            elif attr == "exists" and vval:
                xml_handle_exists(value, elemt_obj, attr, vval)
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
                xml_handle_comment(obj, line_number, line_loc, elemt_obj)
            elif ele_type == "field" and attr == "unit":
                xml_handle_units(elemt_obj, vval)
                xml_handle_comment(obj, line_number, line_loc, elemt_obj)
            elif ele_type == "field" and attr == "dim":
                xml_handle_dim(obj, value)
            elif attr in allowed_attr and not isinstance(vval, dict) and vval:
                validate_field_attribute_and_value(attr, vval, allowed_attr, value)
                elemt_obj.set(attr, check_for_mapping_char_other(vval))
                rm_key_list.append(attr)
                rm_key_list.append(line_number)
                xml_handle_comment(obj, line_number, line_loc, elemt_obj)

        for key in rm_key_list:
            del value[key]
        # Check for skipped attributes
        check_for_skipped_attributes(ele_type, value, allowed_attr, verbose)

    if isinstance(value, dict) and value != {}:
        recursive_build(elemt_obj, value, verbose)


def xml_handle_comment(
    obj: ET._Element,
    line_annotation: str,
    line_loc_no: int,
    xml_ele: ET._Element = None,
    is_def_cmnt: bool = False,
):
    """Handle comment.

        Add xml comment: check for comments searched by 'line_annotation'
    (e.g. __line__data) and line_loc_no (e.g. 30). It
    does following tasks:

    1. Rearrange comment elements of xml_ele and xml_ele where comment comes first.
    2. Append comment element when no xml_ele found as general comments.
    """

    line_info = (line_annotation, int(line_loc_no))
    if line_info in COMMENT_BLOCKS:  # noqa: F821
        cmnt = COMMENT_BLOCKS.get_comment_by_line_info(line_info)  # noqa: F821
        cmnt_text = cmnt.get_comment_text_list()

        # Check comment for definition element and return
        if is_def_cmnt:
            return cmnt_text
        if xml_ele is not None:
            obj.remove(xml_ele)
            for string in cmnt_text:
                # Format comment string to preserve text nxdl to yaml and vice versa
                string = format_nxdl_doc(string)
                obj.append(ET.Comment(string))
            obj.append(xml_ele)
        elif not is_def_cmnt and xml_ele is None:
            for string in cmnt_text:
                obj.append(ET.Comment(string))

    # The searched comment is not related with definition element
    return []


def recursive_build(obj, dct, verbose):
    """Walk through nested dictionary.
    Parameters:
    -----------
    obj : ET.Element
        Obj is the current node of the XML tree where we want to append to.
    dct : dict
     dct is the nested python dictionary which represents the content of a child and
     its successors.

    Note: NXDL fields may contain attributes but trigger no recursion so attributes are leafs.
    """
    for keyword, value in iter(dct.items()):
        if "__line__" in keyword:
            continue
        line_number = f"__line__{keyword}"
        line_loc = dct[line_number]
        keyword_name, keyword_type = nx_name_type_resolving(keyword)
        check_keyword_variable(verbose, dct, keyword, value)
        if verbose:
            print(f"keyword_name:{keyword_name} keyword_type {keyword_type}\n")

        if keyword[-6:] == "(link)":
            xml_handle_link(dct, obj, keyword, value, verbose)
        elif keyword[-8:] == "(choice)":
            xml_handle_choice(dct, obj, keyword, value)
        # The below xml_symbol clause is for the symbols that come ubde filed or attributes
        # Root level symbols has been inside nyaml2nxdl()
        elif keyword_type == "" and keyword_name == "symbols":
            xml_handle_symbols(dct, obj, keyword, value)

        elif re.match(r"NX[a-zA-Z].*", keyword_type) is not None:
            elem_type = "group"
            # we can be sure we need to instantiate a new group
            xml_handle_fields_or_group(
                dct,
                obj,
                keyword,
                value,
                elem_type,
                YAML_GROUP_ATTRIBUTES,
                verbose=False,
            )

        elif keyword_name[0:2] == NX_ATTR_IDNT:  # check if obj qualifies
            xml_handle_attributes(dct, obj, keyword, value, verbose)
        elif keyword == "doc":
            xml_handle_doc(obj, value, line_number, line_loc)
        elif keyword == NX_UNIT_IDNT:
            xml_handle_units(obj, value)
        elif keyword == "enumeration":
            xml_handle_enumeration(dct, obj, keyword, value, verbose)
        elif keyword == "dimensions":
            xml_handle_dimensions(dct, obj, keyword, value)
        elif keyword == "dim":
            xml_handle_dim(obj, value)
        elif keyword == "exists":
            xml_handle_exists(dct, obj, keyword, value)
        # Handles fileds e.g. AXISNAME
        elif keyword_name != "" and "__line__" not in keyword_name:
            elem_type = "field"
            xml_handle_fields_or_group(
                dct,
                obj,
                keyword,
                value,
                elem_type,
                YAML_FIELD_ATTRIBUTES,
                verbose=False,
            )
        else:
            raise ValueError(
                f"An unknown type of element {keyword} has been found which is "
                f"not be able to be resolved. Check around line {dct[line_number]}"
            )


def extend_doc_type(doc_type, new_component, comment=False):
    """Extend doc type for etree.tostring function

    Extend doc type to build DOM and process instruction comments.
    """
    start_sym = "<?"
    end_sym = "?>"
    if comment:
        start_sym = "<!--\n"
        end_sym = "-->"
    return doc_type + "\n" + start_sym + new_component + end_sym


def pretty_print_xml(xml_root, output_xml, def_comments=None):
    """Print in nxdl.xml file.

    Print better human-readable indented and formatted xml file using
    built-in libraries and preceding XML processing instruction
    """
    # Handle DOM as doc_type
    doc_type = '<?xml-stylesheet type="text/xsl" href="nxdlformat.xsl"?>'

    if DOM_COMMENT:
        doc_type = extend_doc_type(doc_type, DOM_COMMENT, comment=True)
    if def_comments:
        for string in def_comments:
            doc_type = extend_doc_type(doc_type, string, comment=True)

    tmp_xml = "tmp.xml"
    ET.indent(xml_root, space=DEPTH_SIZE)
    xml_string = ET.tostring(
        xml_root,
        pretty_print=True,
        encoding="UTF-8",
        xml_declaration=True,
        doctype=doc_type,
    )
    with open(tmp_xml, "wb") as file_tmp:
        file_tmp.write(xml_string)
    flag = False
    with open(tmp_xml, "r", encoding="utf-8") as file_out:
        with open(output_xml, "w", encoding="utf-8") as file_out_mod:
            for i in file_out.readlines():
                i = unquote(i)
                if "<doc>" not in i and "</doc>" not in i and flag is False:
                    file_out_mod.write(i)
                elif "<doc>" in i and "</doc>" in i:
                    file_out_mod.write(i)
                elif "<doc>" in i and "</doc>" not in i:
                    flag = True
                    white_spaces = len(i) - len(i.lstrip())
                    file_out_mod.write(i)
                elif "<doc>" not in i and "</doc>" not in i and flag is True:
                    file_out_mod.write((white_spaces + 5) * " " + i)
                elif "<doc>" not in i and "</doc>" in i and flag is True:
                    file_out_mod.write(white_spaces * " " + i)
                    flag = False
    tmp_xml_path = pathlib.Path(tmp_xml)
    pathlib.Path.unlink(tmp_xml_path)


# pylint: disable=too-many-statements
def nyaml2nxdl(input_file: str, out_file, verbose: bool):
    """
    Main of the nyaml2nxdl converter, creates XML tree, namespace and
    schema, definitions then evaluates a nested dictionary of groups recursively and
    fields or (their) attributes as children of the groups
    """
    def_attributes = [
        "deprecated",
        "ignoreExtraGroups",
        "category",
        "type",
        "ignoreExtraFields",
        "ignoreExtraAttributes",
        "restricts",
    ]
    yml_appdef = yml_reader(input_file)
    def_cmnt_text = []
    if verbose:
        print(f"input-file: {input_file}\n")
        print("application/base contains the following root-level entries:\n")
        print(str(yml_appdef.keys()))
    # etree does not allow to set namespace-map after root creation
    # So, mimic a nsmap and fill it later as dict has hash property
    nsmap = {
        None: "http://definition.nexusformat.org/nxdl/3.1",
    }
    xml_root = ET.Element("definition", attrib={}, nsmap=nsmap)
    assert (
        "category" in yml_appdef.keys()
    ), "Required root-level keyword category is missing!"
    assert yml_appdef["category"] in [
        "application",
        "base",
    ], "Only \
application and base are valid categories!"
    assert "doc" in yml_appdef.keys(), "Required root-level keyword doc is missing!"

    name_extends = ""
    yml_appdef_copy = yml_appdef.copy()
    for kkey, vvalue in yml_appdef_copy.items():
        if "__line__" in kkey:
            continue
        line_number = f"__line__{kkey}"
        line_loc_no = yml_appdef[line_number]
        if not isinstance(vvalue, dict) and kkey in def_attributes:
            xml_root.set(kkey, str(vvalue) or "")
            cmnt_text = xml_handle_comment(
                xml_root, line_number, line_loc_no, is_def_cmnt=True
            )
            def_cmnt_text += cmnt_text

            del yml_appdef[line_number]
            del yml_appdef[kkey]
        # Taking care of name and extends
        elif "NX" in kkey:
            # Taking the attribute order but the correct value will be stored later
            # check for name first or type first if (NXobject)NXname then type first
            l_bracket_ind = kkey.rfind("(")
            r_bracket_ind = kkey.rfind(")")
            if l_bracket_ind == 0:
                extend = kkey[1:r_bracket_ind]
                name = kkey[r_bracket_ind + 1 :]
                xml_root.set("extends", extend)
                xml_root.set("name", name)
            elif l_bracket_ind > 0:
                name = kkey[0:l_bracket_ind]
                extend = kkey[l_bracket_ind + 1 : r_bracket_ind]
                xml_root.set("name", name)
                xml_root.set("extends", extend)
            else:
                name = kkey
                xml_root.set("name", name)
                xml_root.set("extends", "NXobject")
            cmnt_text = xml_handle_comment(
                xml_root, line_number, line_loc_no, is_def_cmnt=True
            )
            def_cmnt_text += cmnt_text if cmnt_text else []

            name_extends = kkey

    if "type" not in xml_root.attrib:
        xml_root.set("type", "group")
    # Taking care of namespaces
    namespaces = {
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
    }
    # Fill nsmap variable here
    nsmap.update(namespaces)  # type: ignore
    xml_root.attrib[
        "{http://www.w3.org/2001/XMLSchema-instance}schemaLocation"
    ] = "http://definition.nexusformat.org/nxdl/3.1 ../nxdl.xsd".replace(" ", "%20")

    # Taking care of Symbols elements
    if "symbols" in yml_appdef.keys():
        xml_handle_symbols(yml_appdef, xml_root, "symbols", yml_appdef["symbols"])

        del yml_appdef["symbols"]
        del yml_appdef["__line__symbols"]
    if isinstance(yml_appdef["doc"], str):
        assert yml_appdef["doc"] != "", "Doc has to be a non-empty string!"
    elif isinstance(yml_appdef["doc"], list):
        assert any(
            yml_appdef["doc"]
        ), "One of the doc elements has to be a non-empty string!"

    line_number = "__line__doc"
    line_loc_no = yml_appdef[line_number]
    xml_handle_doc(xml_root, yml_appdef["doc"], line_number, line_loc_no)

    del yml_appdef["doc"]

    root_keys = 0
    for key in yml_appdef.keys():
        if "__line__" not in key:
            root_keys += 1
            extra_key = key

    assert root_keys == 1, (
        f"Accepting at most keywords: category, doc, symbols, and NX... "
        f"at root-level! check key at root level {extra_key}"
    )

    assert (
        "NX" in name_extends and len(name_extends) > 2
    ), "NX \
keyword has an invalid pattern, or is too short!"
    # Taking care if definition has empty content
    if yml_appdef[name_extends]:
        recursive_build(xml_root, yml_appdef[name_extends], verbose)
    # Taking care of comments that comes at the end of file that is might not be intended for
    # any nxdl elements.
    if COMMENT_BLOCKS[-1].has_post_comment:  # noqa: F821
        post_comment = COMMENT_BLOCKS[-1]  # noqa: F821
        (lin_annot, line_loc) = post_comment.get_line_info()
        xml_handle_comment(xml_root, lin_annot, line_loc)

    # Note: Just to keep the functionality if we need this later.
    default_attr = False
    if default_attr:
        check_for_default_attribute_and_value(xml_root)
    pretty_print_xml(xml_root, out_file, def_cmnt_text)
    if verbose:
        print("Parsed YAML to NXDL successfully\n")
